@@ -809,20 +809,20 @@ TruncateCLOG(TransactionId oldestXact)
 	oldestXact -= oldestXact % CLOG_XACTS_PER_SEGMENT;
 	cutoffPage = TransactionIdToPage(oldestXact);
 
-#ifdef USE_REPLICATION
-	/* Perform a forced CHECKPOINT */
-	CreateCheckPoint(false, true);
-#endif  /* USE_REPLICATION */
 	if (!ScanCLOGDirectory(cutoffPage, false))
-		return;					/* nothing to remove */
-
-	/* Perform a forced CHECKPOINT */
-	CreateCheckPoint(false, true);
-
-#ifndef USR_REPLICATION
-	/* Perform a forced CHECKPOINT */
-	CreateCheckPoint(false, true);
+#ifdef USE_REPLICATION
+	 {
+		/* Perform a forced CHECKPOINT */
+		CreateCheckPoint(false, true);
 #endif  /* USE_REPLICATION */
+		return;					/* nothing to remove */
+#ifdef USE_REPLICATION
+	 }
+#endif  /* USE_REPLICATION */
+
+	/* Perform a forced CHECKPOINT */
+	CreateCheckPoint(false, true);
+
 	/*
 	 * Scan CLOG shared memory and remove any pages preceding the cutoff
 	 * page, to ensure we won't rewrite them later.  (Any dirty pages
