@@ -65,46 +65,15 @@
 #include "utils/guc.h"
 #include "parser/parser.h"
 #include "access/xact.h"
+#include "storage/proc.h"
 #include "replicate.h"
 
+PGR_ReplicationLog_Info ReplicationLog_Info;
+
 /*--------------------------------------
  * PROTOTYPE DECLARATION
  *--------------------------------------
  */
-int PGR_Init_Replicate_Server_Data(void);
-int PGR_Set_Replicate_Server_Socket(void);
-char * PGR_Send_Replicate_Command(char * query_string, int query_len, char cmdSts ,char cmdType);
-bool PGR_Is_Replicated_Command(char * query);
-int Xlog_Check_Replicate(int operation);
-int PGR_Replicate_Function_Call(void);
-void PGR_delete_shm(void);
-int PGR_replication(char * query_string, CommandDest dest, MemoryContext parse_context, const char * commandTag);
-ReplicateServerInfo * PGR_get_replicate_server_info(void);
-bool PGR_Is_System_Command(char * query);
-int PGR_Call_System_Command(char * command);
-int PGR_GetTimeOfDay(struct timeval *tp, struct timezone *tpz);
-long PGR_Random(void);
-int PGR_Set_Current_Time(char * sec, char * usec);
-int PGR_Send_Copy(CopyData * copy,int end);
-CopyData * PGR_Set_Copy_Data(CopyData * copy, char * str, int len, int end);
-char * PGR_scan_terminate( char * str);
-bool PGR_Is_Stand_Alone(void);
-void PGR_Send_Message_To_Frontend(char * msg);
-void PGR_Notice_Transaction_Query_Done(void);
-void PGR_Notice_Transaction_Query_Aborted(void);
-int PGR_Notice_Conflict(void);
-int PGR_Recv_Trigger (int user_timeout);
-bool PGR_Check_DeadLock(LOCKMETHODTABLE *lockMethodTable, LOCKMODE lockmode, LOCK *lock, PROCLOCK *holder);
-void PGR_Set_Replication_Server_Status( ReplicateServerInfo * sp, int status);
-int PGR_Is_Skip_Replication(char * query);
-bool PGR_Did_Commit_Transaction(void);
-bool PGR_Set_Transaction_Mode(bool mode,const char * commandTag);
-char * PGR_Remove_Comment(char * str);
-void PGR_Force_Replicate_Query(void);
-void PGR_Notice_DeadLock(void);
-void PGR_Set_Cluster_Status(int status);
-int PGR_Get_Cluster_Status(void);
-
 static int set_command_args(char argv[PGR_CMD_ARG_NUM][256],char *str);
 static bool is_same_replication_server(ReplicateServerInfo * sp1, ReplicateServerInfo * sp2 );
 static ReplicateServerInfo * search_new_replication_server ( ReplicateServerInfo * sp , int socket_type );
@@ -116,7 +85,7 @@ static void clear_recv_queue(int sock);
 static int send_replicate_packet(int sock,ReplicateHeader * header, char * query_string);
 static int get_replicate_server_socket ( ReplicateServerInfo * sp , int socket_type );
 static bool is_copy_from(char * query);
-static int is_session_authorization(char * query);
+//static int is_session_authorization(char * query);
 static int get_words( char words[MAX_WORDS][MAX_WORD_LETTERS] ,char * string,int length,int upper);
 static int get_table_name(char * table_name, char * query, int position );
 static bool is_not_replication_query(char * query_string, int query_len, char cmdType);
@@ -132,11 +101,17 @@ static int change_replication_server(char * hostname,char * port, char * recover
 static int get_new_replication_socket( ReplicateServerInfo * base, ReplicateServerInfo * sp, int socket_type);
 static char * get_hostName(char * str);
 static void set_response_mode(char * mode);
+static void PGR_Set_Current_Replication_Query_ID(char *id);
 #ifdef CONTROL_LOCK_CONFLICT
 static int wait_lock_answer(void);
 static int read_trigger(char * result, int buf_size);
 #endif /* CONTROL_LOCK_CONFLICT */
 
+
+static unsigned int get_next_request_id(void);
+static bool is_this_query_replicated(char * id);
+static int set_replication_id(char * id);
+
 extern ssize_t secure_read(Port *, void *, size_t);
 /*--------------------------------------------------------------------
  * SYMBOL
@@ -158,6 +133,9 @@ PGR_Init_Replicate_Server_Data(void)
 	PGR_Not_Replicate_Type * nrp;
 	ConfDataType * conf;
 	int rec_no,cnt;
+	unsigned int ip;
+	char HostName[128];
+
 
 	if (ConfData_Top == (ConfDataType *)NULL)
 	{
@@ -223,7 +201,7 @@ PGR_Init_Replicate_Server_Data(void)
 				(sp + rec_no)->sock = -1;
 				if ((sp + rec_no)->useFlag != DATA_USE)
 				{
-					PGR_Set_Replication_Server_Status((sp+rec_no), DATA_USE);
+					PGR_Set_Replication_Server_Status((sp+rec_no), DATA_INIT);
 				}
 				memset((sp + rec_no + 1)->hostName,0,sizeof(sp->hostName));
 				(sp + rec_no + 1)->useFlag = DATA_END;
@@ -235,7 +213,7 @@ PGR_Init_Replicate_Server_Data(void)
 				(sp + rec_no)->recoveryPortNumber = atoi(conf->value);
 				if ((sp + rec_no)->useFlag != DATA_USE)
 				{
-					PGR_Set_Replication_Server_Status((sp+rec_no), DATA_USE);
+					PGR_Set_Replication_Server_Status((sp+rec_no), DATA_INIT);
 				}
 				memset((sp + rec_no + 1)->hostName,0,sizeof(sp->hostName));
 				(sp + rec_no + 1)->useFlag = DATA_END;
@@ -291,6 +269,13 @@ PGR_Init_Replicate_Server_Data(void)
 			memset(RsyncOption,0,str_size + 1);
 			memcpy(RsyncOption,conf->value,str_size);
 		}
+		if (!strcmp(conf->key,RSYNC_COMPRESS_TAG))
+		{
+			if (!strcmp(conf->value, "yes"))
+				RsyncCompress = true;
+			else if (!strcmp(conf->value, "no"))
+				RsyncCompress = false;
+		}
 		if (!strcmp(conf->key,STAND_ALONE_TAG))
 		{
 			PGR_Stand_Alone = (PGR_Stand_Alone_Type*)malloc(sizeof(PGR_Stand_Alone_Type));
@@ -341,10 +326,18 @@ PGR_Init_Replicate_Server_Data(void)
 		return STATUS_ERROR;
 	}
 	memset(PGRSelfHostName,0,HOSTNAME_MAX_LENGTH);
-	if (gethostname(PGRSelfHostName,HOSTNAME_MAX_LENGTH) < 0)
+	if (gethostname(HostName,HOSTNAME_MAX_LENGTH) < 0)
 	{
 		return STATUS_ERROR;
 	}
+	ip=PGRget_ip_by_name(HostName);
+
+        sprintf(PGRSelfHostName,
+					 "%d.%d.%d.%d",
+					 (ip      ) & 0xff ,
+					 (ip >>  8) & 0xff ,
+					 (ip >> 16) & 0xff ,
+					 (ip >> 24) & 0xff );
 
 	return	STATUS_OK;
 }
@@ -669,7 +662,7 @@ is_not_replication_query(char * query_string, int query_len, char cmdType)
 char *
 PGR_Send_Replicate_Command(char * query_string, int query_len, char cmdSts ,char cmdType)
 {
-	int sock = 0;
+	int sock = -1;
 	int cnt = 0;
 	ReplicateHeader header;
 	char * serverName = NULL;
@@ -681,6 +674,8 @@ PGR_Send_Replicate_Command(char * query_string, int query_len, char cmdSts ,char
 	char argv[ PGR_CMD_ARG_NUM ][256];
 	int argc = 0;
 	int func_no = 0;
+	int check_flag =0;
+
 
 	/*
 	 * check query string
@@ -730,7 +725,7 @@ PGR_Send_Replicate_Command(char * query_string, int query_len, char cmdSts ,char
 	serverName = sp->hostName;
 	portNumber = (int)sp->portNumber;
 	memset(&header,0,sizeof(ReplicateHeader));
-	header.cmdSys = CMD_SYS_REPLICATE;
+
 	header.cmdSts = cmdSts;
 	header.cmdType = cmdType;
 	header.port = htons(PostPortNumber);
@@ -738,6 +733,9 @@ PGR_Send_Replicate_Command(char * query_string, int query_len, char cmdSts ,char
 	header.query_size = htonl(query_len); 
 	strncpy(header.dbName ,(char *)(MyProcPort->database),sizeof(header.dbName));
 	strncpy(header.userName , (char *)(MyProcPort->user),sizeof(header.userName));
+	header.request_id = htonl(get_next_request_id());
+	header.rlog = 0;
+
 	if (PGRSelfHostName != NULL)
 	{
 		strncpy(header.from_host, PGRSelfHostName, HOSTNAME_MAX_LENGTH);
@@ -746,11 +744,98 @@ PGR_Send_Replicate_Command(char * query_string, int query_len, char cmdSts ,char
 	base = sp;
 	PGR_Sock_To_Replication_Server = sock;
 
-retry_send_replicate_packet:
+retry_send_prereplicate_packet:
 
-	clear_recv_queue(sock); 
+	//	clear_recv_queue(sock); 
 	memset(result,0,PGR_MESSAGE_BUFSIZE + 4);
 	cnt = 0;
+	header.cmdSys=CMD_SYS_PREREPLICATE;
+
+	while (send_replicate_packet(sock,&header,query_string) != STATUS_OK)
+	{
+		cnt++;
+		if (cnt >= MAX_RETRY_TIMES )
+		{
+			sock = get_new_replication_socket( base, sp, socket_type);
+			if (sock < 0)
+			{
+				if (Debug_pretty_print)
+					elog(DEBUG1,"all replication servers may be down");
+				PGR_Stand_Alone->is_stand_alone = true;
+				if (cmdSts == CMD_STS_TRANSACTION )
+				{
+					strcpy(result,PGR_REPLICATION_ABORT_MSG);
+					return result;
+				}
+				free(result);
+				result = NULL;
+				return NULL;
+				
+			}
+			if(cmdSts==CMD_STS_TRANSACTION ) 
+			{
+				if (cmdType!=CMD_TYPE_BEGIN 
+				    && cmdType!=CMD_TYPE_ROLLBACK)  
+				{							
+					elog(ERROR,"replicate server down during replicating transaction. aborted.");
+					free(result);
+					return NULL;
+				}
+			}					 
+			PGR_Sock_To_Replication_Server = sock;
+			cnt = 0;
+		}
+	}
+
+	memset(result,0,PGR_MESSAGE_BUFSIZE);
+	if (recv_replicate_result(sock,result,0) < 0)
+	{
+
+			sock = get_new_replication_socket( base, sp, socket_type);
+			if (sock < 0)
+			{
+					if (Debug_pretty_print)
+							elog(DEBUG1,"all replication servers may be down");
+					PGR_Stand_Alone->is_stand_alone = true;
+
+					if (cmdSts == CMD_STS_TRANSACTION )
+					{
+							strcpy(result,PGR_REPLICATION_ABORT_MSG);
+							return result;
+					}
+					if(result!=NULL) {
+							free(result);
+							result = NULL;
+					}
+					return NULL;
+			}
+			PGR_Sock_To_Replication_Server = sock;
+			/* replication server should be down */
+
+			if(cmdSts==CMD_STS_TRANSACTION 
+			   && (cmdType!=CMD_TYPE_BEGIN && cmdType!=CMD_TYPE_ROLLBACK))
+			{
+					elog(ERROR,"replicate server down during replicating transaction. aborted.");
+					free(result);
+					return NULL;
+			}
+
+			goto retry_send_prereplicate_packet;
+	}
+
+
+	argc = set_command_args(argv,result);
+	func_no=atoi(argv[0]);
+	if(func_no==0) {
+			/* this server is not primary replicate server*/
+			sock=-1;
+			goto retry_send_prereplicate_packet;
+	}
+retry_send_replicate_packet:
+
+	memset(result,0,PGR_MESSAGE_BUFSIZE + 4);
+	cnt = 0;
+	header.cmdSys = CMD_SYS_REPLICATE;
 	while (send_replicate_packet(sock,&header,query_string) != STATUS_OK)
 	{
 		if (cnt > MAX_RETRY_TIMES )
@@ -759,20 +844,22 @@ retry_send_replicate_packet:
 			if (sock < 0)
 			{
 				if (Debug_pretty_print)
-					elog(DEBUG1,"all replication servers may be down");
+				  elog(DEBUG1,"all replication servers may be down");
 				PGR_Stand_Alone->is_stand_alone = true;
+				if (cmdSts == CMD_STS_TRANSACTION )
+				{
+				        strcpy(result,PGR_REPLICATION_ABORT_MSG);
+				        return result;
+				}
 				free(result);
 				result = NULL;
 				return NULL;
+
 			}
 			PGR_Sock_To_Replication_Server = sock;
+			header.rlog = CONNECTION_SUSPENDED_TYPE;
 			cnt = 0;
 		}
-		if (cmdSts == CMD_STS_TRANSACTION )
-		{
-			strcpy(result,PGR_REPLICATION_ABORT_MSG);
-			return result;
-		}
 		cnt ++;
 	}
 
@@ -786,16 +873,21 @@ retry_send_replicate_packet:
 			if (Debug_pretty_print)
 				elog(DEBUG1,"all replication servers may be down");
 			PGR_Stand_Alone->is_stand_alone = true;
-			free(result);
-			result = NULL;
+
+			if (cmdSts == CMD_STS_TRANSACTION )
+			{
+			        strcpy(result,PGR_REPLICATION_ABORT_MSG);
+				return result;
+			}
+			if(result!=NULL) {
+			        free(result);
+			        result = NULL;
+			}
 			return NULL;
 		}
 		PGR_Sock_To_Replication_Server = sock;
-		if (cmdSts == CMD_STS_TRANSACTION )
-		{
-			strcpy(result,PGR_REPLICATION_ABORT_MSG);
-			return result;
-		}
+		header.rlog = CONNECTION_SUSPENDED_TYPE;
+
 		goto retry_send_replicate_packet;
 	}
 
@@ -806,21 +898,153 @@ retry_send_replicate_packet:
 		if (func_no == PGR_SET_CURRENT_TIME_FUNC_NO)
 		{
 			PGR_Set_Current_Time(argv[1],argv[2]);
-			set_response_mode(argv[3]);
+			set_replication_id(argv[3]);
+			set_response_mode(argv[4]);
+         		PGR_Set_Current_Replication_Query_ID(argv[5]);
 		}
 		else if (func_no == PGR_NOTICE_DEADLOCK_DETECTION_FUNC_NO)
 		{
 			memset(result,0,PGR_MESSAGE_BUFSIZE);
 			strcpy(result,PGR_DEADLOCK_DETECTION_MSG);
 		}
+                else if (func_no == PGR_SET_CURRENT_REPLICATION_QUERY_ID_NO) 
+		{
+		  PGR_Set_Current_Replication_Query_ID(argv[1]);
+		}  
+		else if (func_no == PGR_QUERY_CONFIRM_ANSWER_FUNC_NO)
+		{
+			check_flag = atoi(argv[1]);
+			if (check_flag == PGR_ALREADY_COMMITTED )
+			{
+				PGR_Set_Current_Time(argv[2],argv[3]);
+				set_replication_id(argv[4]);
+			}
+			else
+			{
+				PGR_Set_Current_Time(argv[1],argv[2]);
+				set_replication_id(argv[3]);
+				/* this query is not replicated */
+				/*
+				free(result);
+				return NULL;
+				*/
+			}
+		}
 	}
 	return result;
 }
 
+
+static int
+set_replication_id(char * id)
+{
+        uint32_t rid=0;
+		uint32_t saved_id;
+		if (id == NULL)
+		{
+				return STATUS_ERROR;
+		}
+
+		rid=(uint32_t)atol(id);
+		if(rid==0)
+				return STATUS_OK;
+
+		needToUpdateReplicateIdOnNextQueryIsDone=true;
+		saved_id=ReplicationLog_Info.PGR_Replicate_ID;
+
+		ReplicationLog_Info.PGR_Replicate_ID =rid;
+
+
+		/*set replicate id in this process */
+
+
+		if (CurrentReplicateServer == NULL)
+		{
+				PGR_get_replicate_server_info();
+		}
+		if (CurrentReplicateServer != NULL)
+		{
+				/* set replicate id in this system */
+				saved_id=CurrentReplicateServer->replicate_id;
+				elog(DEBUG1, "replication id set from %d to %d", saved_id, rid); 
+
+				CurrentReplicateServer->replicate_id = (uint32_t)(atol(id));
+		}
+	
+		return STATUS_OK;
+}
+
+
+static unsigned int
+get_next_request_id(void)
+{
+	if (ReplicationLog_Info.PGR_Request_ID +1 < PGR_MAX_COUNTER)
+	{
+		ReplicationLog_Info.PGR_Request_ID ++;
+	}
+	else
+	{
+		ReplicationLog_Info.PGR_Request_ID = 0;
+	}
+	return ReplicationLog_Info.PGR_Request_ID ;
+		
+}
+
+static bool
+is_this_query_replicated(char * id)
+{
+	uint32_t replicate_id = 0;
+	uint32_t saved_id = 0;
+	int32_t diff=0;
+	ReplicateServerInfo * replicate_server_info = NULL;
+
+	if (id == NULL)
+	{
+		return false;
+	}
+	replicate_id = (uint32_t)atol(id);
+	elog(DEBUG1, "check for replication id , input=%u", replicate_id);
+
+	if (CurrentReplicateServer == NULL)
+	{
+		PGR_get_replicate_server_info();
+	}
+
+	if (CurrentReplicateServer != NULL)
+	{
+		replicate_server_info = CurrentReplicateServer;
+	}
+        else if (LastReplicateServer != NULL)
+	{	 
+		replicate_server_info = LastReplicateServer;
+	}
+	if (replicate_server_info != NULL)
+	{
+
+	        saved_id=replicate_server_info->replicate_id;
+		saved_id = saved_id < ReplicationLog_Info.PGR_Replicate_ID
+		  ? ReplicationLog_Info.PGR_Replicate_ID
+		  : saved_id;
+
+		elog(DEBUG1, "check for replication id , now=%u", saved_id);
+		/* check replicate_id < saved_id logically 
+		 * 
+		 * see also:
+		 *  backend/transam/transam.c#TransactionIdPrecedes
+		 */
+
+	        diff = (int32) (saved_id-replicate_id);
+		return (diff > 0);
+	}
+	elog(DEBUG1, "check for replication id check failed. no replication server");
+	return false;
+}
+
+
 static int
 get_new_replication_socket( ReplicateServerInfo * base, ReplicateServerInfo * sp, int socket_type)
 {
-	int sock = 0;
+	int sock = -1;
 
 	if (( base == NULL) ||
 		( sp == NULL))
@@ -853,36 +1077,32 @@ recv_replicate_result(int sock,char * result,int user_timeout)
 	{
 		return -1;
 	}
-	if (user_timeout == 0)
-	{
-		timeout.tv_sec = PGR_RECV_TIMEOUT * 3;
-	}
-	else
-	{
-		timeout.tv_sec = user_timeout;
-	}
-	timeout.tv_usec = 0;
 
 	/*
 	 * Wait for something to happen.
 	 */
-	rtn = 1;
-	while (rtn)
+	for (;;)
 	{
+		if (user_timeout == 0)
+			timeout.tv_sec = PGR_RECV_TIMEOUT * 3;
+		else
+			timeout.tv_sec = user_timeout;
+
+		timeout.tv_usec = 0;
+
 		FD_ZERO(&rmask);
 		FD_SET(sock,&rmask);
 		rtn = select(sock+1, &rmask, (fd_set *)NULL, (fd_set *)NULL, &timeout);
-		if ((rtn > 0) && (FD_ISSET(sock, &rmask)))
+		if (rtn < 0)
+		{
+			if (errno != EINTR)
+				return -1;
+		}
+
+		else if ((rtn > 0) && (FD_ISSET(sock, &rmask)))
 		{
 			return (recv_message(sock, result,0));
 		}
-		else
-		{
-			if (errno != EINTR) 
-			{
-				return -1;
-			}
-		}
 	}
 	return -1;
 }
@@ -900,71 +1120,25 @@ recv_message(int sock,char * buf,int flag)
 	for (;;)
 	{
 		r = recv(sock,read_ptr + read_size ,PGR_MESSAGE_BUFSIZE - read_size, flag); 
-		if (r < 0)
-		{
-			if (errno == EINTR)
-			{
-				usleep(PGR_RECV_WAIT_MSEC);
+		if (r < 0) {
+			if (errno == EINTR || errno == EAGAIN) {
 				continue;
+			} else {
+			  elog(DEBUG1, "recv_message():recv failed");
+			  return -1;
 			}
-#ifdef EAGAIN
-			else if (errno == EAGAIN)
-			{
-				usleep(PGR_RECV_WAIT_MSEC);
-				continue;
-			}
-#endif /* EAGAIN */
-#ifdef ECONNREFUSED
-			else if (errno == ECONNREFUSED)
-			{
-				usleep(PGR_RECV_WAIT_MSEC);
-				continue;
-			}
-#endif /* ECONNREFUSED */
-#ifdef ENOTCONN
-			else if (errno == ENOTCONN)
-			{
-				usleep(PGR_RECV_WAIT_MSEC);
-				continue;
-			}
-#endif /* ENOTCONN */
-			if (cnt < PGR_RECV_RETRY_CNT )
-			{
-				cnt ++;
-				usleep(PGR_RECV_WAIT_MSEC);
-				continue;
-			}
-			else
-			{
-				if (read_size == 0)
-				{
-					return -1;
-				}
-			}
-		}
-		if (r > 0)
-		{
+		} else if (r == 0) {
+			elog(DEBUG1, "recv_message():unexpected EOF");
+			return -1;
+		} else /*if (r > 0)*/ {
 			read_size += r;
 			if (read_size == PGR_MESSAGE_BUFSIZE)
 			{
 				return read_size;
 			}
 		}
-		if ((r == 0) && (read_size == 0) && (cnt > 0))
-		{
-			return -1;
-		}
-		if (cnt < PGR_RECV_RETRY_CNT )
-		{
-			cnt ++;
-			usleep(PGR_RECV_WAIT_MSEC);
-			continue;
-		}
-		else
-		{
-			return -1;
-		}
 	}
+	return -1;
 }
 
 static void
@@ -975,32 +1149,27 @@ clear_recv_queue(int sock)
 	char buf[PGR_MESSAGE_BUFSIZE];
 	int rtn = 0;
 
-	timeout.tv_sec = 0;
-	timeout.tv_usec = 0;
-
-	FD_ZERO(&rmask);
-	FD_SET(sock,&rmask);
-
 	memset(buf,0,sizeof(buf));
 	rtn = 1;
-	while (rtn)
+	for (;;)
 	{
+		timeout.tv_sec = 10;
+		timeout.tv_usec = 0;
+
 		FD_ZERO(&rmask);
 		FD_SET(sock,&rmask);
 		rtn = select(sock+1, &rmask, (fd_set *)NULL, (fd_set *)NULL, &timeout);
-		if ((rtn > 0) && (FD_ISSET(sock, &rmask)))
+		if (rtn < 0)
+		{
+			if (errno != EINTR)
+				break;
+		}
+
+		else if ((rtn > 0) && (FD_ISSET(sock, &rmask)))
 		{
 			if (recv_message(sock, buf,0) <= 0)
 				break;
 		}
-		else if (errno == EINTR)
-		{
-			continue;
-		}
-		else
-		{
-			break;
-		}
 	}
 }
 
@@ -1008,7 +1177,6 @@ static int
 send_replicate_packet(int sock,ReplicateHeader * header, char * query_string)
 {
 	int s = 0;
-	int cnt = 0;
 	char * send_ptr = NULL;
 	char * buf = NULL;
 	int send_size = 0;
@@ -1042,87 +1210,55 @@ send_replicate_packet(int sock,ReplicateHeader * header, char * query_string)
 	}
 	send_ptr = buf;
 
-	timeout.tv_sec = PGR_SEND_TIMEOUT;
-	timeout.tv_usec = 0;
-
 	/*
 	 * Wait for something to happen.
 	 */
 	rtn = 1;
-	while (rtn)
+	for (;;)
 	{
+		timeout.tv_sec = PGR_SEND_TIMEOUT;
+		timeout.tv_usec = 0;
+
 		FD_ZERO(&wmask);
 		FD_SET(sock,&wmask);
 		rtn = select(sock+1, (fd_set *)NULL, &wmask, (fd_set *)NULL, &timeout);
-		if (rtn && FD_ISSET(sock, &wmask))
+		if (rtn < 0)
 		{
-			cnt = 0;
-			send_size = 0;
-			for (;;)
+			if (errno == EINTR)
+				continue;
+			else
 			{
-				/* s = send(sock,send_ptr + send_size,buf_size - send_size ,MSG_DONTWAIT); */
-				s = send(sock,send_ptr + send_size,buf_size - send_size ,0);
-				if (s < 0){
-					if (errno == EINTR)
-					{
-						usleep(PGR_SEND_WAIT_MSEC);
-						continue;
-					}
-	#ifdef EAGAIN
-					if (errno == EAGAIN)
-					{
-						usleep(PGR_SEND_WAIT_MSEC);
-						continue;
-					}
-	#endif /* EAGAIN */
-	#ifdef EWOULDBLOCK
-					if (errno == EWOULDBLOCK )
-					{
-						usleep(PGR_SEND_WAIT_MSEC);
-						continue;
-					}
-	#endif /* EWOULDBLOCK */
-	#ifdef ENOBUF
-					if (errno == ENOBUF )
-					{
-						usleep(PGR_SEND_WAIT_MSEC);
-						continue;
-					}
-	#endif /* ENOBUF */
-	#ifdef ENOMEM
-					if (errno == ENOMEM )
-					{
-						usleep(PGR_SEND_WAIT_MSEC);
-						continue;
-					}
-	#endif /* ENOMEM */
-					if (cnt < PGR_SEND_RETRY_CNT )
-					{
-						cnt ++;
-						usleep(PGR_SEND_WAIT_MSEC);
-						continue;
-					}
-					else
-					{
-						if (send_size == 0)
-						{
-							free(buf);
-							buf = NULL;
-							return STATUS_ERROR;
-						}
-					}
-				}
-				if (s > 0)
+				elog(DEBUG1, "send_replicate_packet():select() failed");
+				return STATUS_ERROR;
+			}
+		}
+		else if (rtn && FD_ISSET(sock, &wmask))
+		{
+
+
+			s = send(sock,send_ptr + send_size,buf_size - send_size ,0);
+			if (s < 0){
+				if (errno == EINTR || errno == EAGAIN)
 				{
-					send_size += s;
-					if (send_size == buf_size)
-					{
-						free(buf);
-						buf = NULL;
-						return STATUS_OK;
-					}
+					continue;
+				}
+				elog(DEBUG1, "send_replicate_packet():send error");
+
+				/* EPIPE || ENCONNREFUSED || ENSOCK || EHOSTUNREACH */
+				return STATUS_ERROR;
+			} else if (s == 0) {
+				free(buf);
+				buf = NULL;
+				elog(DEBUG1, "send_replicate_packet():unexpected EOF");
+				return STATUS_ERROR;
+			} else /*if (s > 0)*/ {
+				send_size += s;
+				if (send_size == buf_size)
+				{
+					free(buf);
+					buf = NULL;
+					return STATUS_OK;
 				}
-				usleep(PGR_SEND_WAIT_MSEC);
 			}
 		}
 	}
@@ -1279,6 +1415,11 @@ PGR_get_replicate_server_info(void)
 	{
 		if (sp->useFlag == DATA_USE )
 		{
+			if (CurrentReplicateServer != NULL)
+			{
+				LastReplicateServer = CurrentReplicateServer;
+				CurrentReplicateServer->replicate_id = LastReplicateServer->replicate_id;
+			}
 			CurrentReplicateServer = sp;
 			return sp;
 		}
@@ -1289,16 +1430,29 @@ PGR_get_replicate_server_info(void)
 	{
 		if (sp->useFlag != DATA_ERR )
 		{
+			if (CurrentReplicateServer != NULL)
+			{
+				LastReplicateServer = CurrentReplicateServer;
+				CurrentReplicateServer->replicate_id = LastReplicateServer-> replicate_id;
+			}
 			CurrentReplicateServer = sp;
+			PGR_Set_Replication_Server_Status(sp, DATA_USE);
 			return sp;
 		}
 		sp++;
 	}
 	PGR_Stand_Alone->is_stand_alone = true;
+	if (CurrentReplicateServer != NULL)
+	{
+	  LastReplicateServer = CurrentReplicateServer;
+	  CurrentReplicateServer->replicate_id = LastReplicateServer-> replicate_id;
+	}
 	CurrentReplicateServer = NULL;
 	return (ReplicateServerInfo *)NULL;
 }
 
+
+
 int
 PGR_Send_Copy(CopyData * copy,int end )
 {
@@ -1414,13 +1568,12 @@ PGR_Set_Copy_Data(CopyData * copy, char *str, int len,int end)
 }
 
 int
-PGR_replication(char * query_string, CommandDest dest, MemoryContext parse_context, const char * commandTag)
+PGR_replication(char * query_string, CommandDest dest, Node *parsetree, const char * commandTag)
 {
 	char *result = NULL;
 	char cmdSts = CMD_STS_OTHER;
 	char cmdType = CMD_TYPE_OTHER;
 	int query_len = 0;
-	int session_type;
 
 	if ((query_string == NULL) ||
 		(commandTag == NULL))
@@ -1451,6 +1604,14 @@ PGR_replication(char * query_string, CommandDest dest, MemoryContext parse_conte
 	else if (!strcmp(commandTag,"RESET")) cmdType = CMD_TYPE_RESET ;
 	else if (!strcmp(commandTag,"START TRANSACTION")) cmdType = CMD_TYPE_BEGIN ;
 
+	/* only "replication_server" statement-name is replicated for SHOW. */
+	/*   see CreateCommandTag() @ backend/tcop/postgres.c      */
+	/*
+	else if (!strcmp(commandTag,"SHOW"))
+	{
+	}
+	*/
+
 	else if (!strcmp(commandTag,"COPY"))
 	{
 		cmdType = CMD_TYPE_COPY ;
@@ -1466,22 +1627,19 @@ PGR_replication(char * query_string, CommandDest dest, MemoryContext parse_conte
 	}
 	else if (!strcmp(commandTag,"SET")) 
 	{
+		VariableSetStmt *stmt = (VariableSetStmt *)parsetree;
 		cmdType = CMD_TYPE_SET;
-		session_type = is_session_authorization(query_string);
-		switch (session_type)
-		{
-			case SESSION_AUTHORIZATION_BEGIN:
-				cmdType = CMD_TYPE_SESSION_AUTHORIZATION_BEGIN ;
-				Session_Authorization_Mode = true;
-				break;
-			case SESSION_AUTHORIZATION_END:
-				cmdType = CMD_TYPE_SESSION_AUTHORIZATION_END ;
-				Session_Authorization_Mode = false;
-				break;
-			default:
-				return STATUS_NOT_REPLICATE;
-				break;
-		}
+
+		if (strcmp(stmt->name, "TRANSACTION ISOLATION LEVEL") &&
+			strcmp(stmt->name, "datestyle") &&
+			strcmp(stmt->name, "autocommit") &&
+			strcmp(stmt->name, "client_encoding") &&
+			strcmp(stmt->name, "password_encryption") &&
+			strcmp(stmt->name, "search_path") &&
+			strcmp(stmt->name, "session_authorization") &&
+			strcmp(stmt->name, "timezone"))
+
+			return STATUS_NOT_REPLICATE;
 	}
 	else if (!strcmp(commandTag,"CREATE TABLE")) 
 	{
@@ -1730,13 +1888,13 @@ add_replication_server(char * hostname,char * port, char * recovery_port)
 		return STATUS_ERROR;
 	}
 	portNumber = atoi(port);
-	recoveryPortNumber = atol(recovery_port);
+	recoveryPortNumber = atoi(recovery_port);
 	cnt = 0;
 	sp = ReplicateServerData;
 	while (sp->useFlag != DATA_END){
 		if((!strncmp(sp->hostName,hostname,sizeof(sp->hostName))) &&
 			(sp->portNumber == portNumber) &&
-			(sp->recoveryPortNumber == RecoveryPortNumber))
+			(sp->recoveryPortNumber == recoveryPortNumber))
 		{
 			if (sp->useFlag != DATA_USE)
 			{
@@ -1751,7 +1909,7 @@ add_replication_server(char * hostname,char * port, char * recovery_port)
 	{
 		memcpy(sp->hostName,hostname,sizeof(sp->hostName));
 		sp->portNumber = portNumber;
-		sp->recoveryPortNumber = RecoveryPortNumber;
+		sp->recoveryPortNumber = recoveryPortNumber;
 		PGR_Set_Replication_Server_Status(sp, DATA_INIT);
 		memset((sp+1),0,sizeof(ReplicateServerInfo));
 		(sp + 1)->useFlag = DATA_END;
@@ -1782,13 +1940,13 @@ change_replication_server(char * hostname,char * port, char * recovery_port)
 		return STATUS_ERROR;
 	}
 	portNumber = atoi(port);
-	recoveryPortNumber = atol(recovery_port);
+	recoveryPortNumber = atoi(recovery_port);
 	cnt = 0;
 	sp = ReplicateServerData;
 	while (sp->useFlag != DATA_END){
 		if((!strcmp(sp->hostName,hostname)) &&
 			(sp->portNumber == portNumber) &&
-			(sp->recoveryPortNumber == RecoveryPortNumber))
+			(sp->recoveryPortNumber == recoveryPortNumber))
 		{
 			PGR_Set_Replication_Server_Status(sp, DATA_USE);
 		}
@@ -1825,6 +1983,12 @@ PGR_Set_Current_Time(char * sec, char * usec)
 	return	STATUS_OK;
 }
 
+static void
+PGR_Set_Current_Replication_Query_ID(char *id) {
+      MyProc->replicationId=atol(id);
+      return;
+}
+
 static void
 set_response_mode(char * mode)
 {
@@ -1888,8 +2052,18 @@ PGR_Call_System_Command(char * command)
 			}
 			else
 			{
+			  /*
+			  if ((atoi(argv[3]) > 0) &&
+				(is_this_query_replicated(argv[3]) == true))
+			    {				
+			      return STATUS_SKIP_QUERY;
+			    }
+			  */
 				PGR_Set_Current_Time(argv[1],argv[2]);
-				set_response_mode(argv[3]);
+				set_replication_id(argv[3]);
+				set_response_mode(argv[4]);
+				PGR_Set_Current_Replication_Query_ID(argv[5]);
+
 			}
 			break;
 		/* add new replication server data */
@@ -1902,6 +2076,22 @@ PGR_Call_System_Command(char * command)
 			hostName = get_hostName(argv[1]);
 			change_replication_server(hostName,argv[2],argv[3]);
 			break;
+        	case PGR_SET_CURRENT_REPLICATION_QUERY_ID_NO:
+		  PGR_Set_Current_Replication_Query_ID(argv[1]);
+			break;
+		case PGR_QUERY_CONFIRM_ANSWER_FUNC_NO:
+			if ((atoi(argv[3]) > 0) &&
+				(is_this_query_replicated(argv[3]) == true))
+			{
+				/* skip this query */
+			  return STATUS_SKIP_QUERY;
+			}
+			else
+			{
+				PGR_Set_Current_Time(argv[1],argv[2]);
+				set_replication_id(argv[3]);
+			}
+			break;
 	}
 	return STATUS_OK;
 }
@@ -1934,7 +2124,7 @@ PGR_Random(void)
 	{
 		if ( ReplicateCurrentTime->use_seed == 0)
 		{
-			srand( ReplicateCurrentTime->tp.tv_sec );
+			srand( ReplicateCurrentTime->tp.tv_usec );
 			ReplicateCurrentTime->use_seed = 1;
 		}
 	}
@@ -2066,7 +2256,8 @@ is_create_temp_table(char * query)
 	}
 	return false;
 }
-	
+
+#if 0
 static int
 is_session_authorization(char * query)
 {
@@ -2115,6 +2306,7 @@ is_session_authorization(char * query)
 	}
 	return rtn;
 }
+#endif
 
 static int
 get_words( char words[MAX_WORDS][MAX_WORD_LETTERS] ,char * string,int length,int upper)
@@ -2292,6 +2484,12 @@ is_select_into_query(char cmdType,char * query)
 		buf = NULL;
 		return true;
 	}
+	if (strstr(buf,"CREATE") != NULL)
+	{
+		free(buf);
+		buf = NULL;
+		return true;
+	}
 	free(buf);
 	buf = NULL;
 	return false;
@@ -2514,7 +2712,6 @@ do_not_replication_command(const char * commandTag)
 		(!strcmp(commandTag,"CLOSE CURSOR")) ||
 		(!strcmp(commandTag,"MOVE")) ||
 		(!strcmp(commandTag,"FETCH")) ||
-		(!strcmp(commandTag,"SHOW")) ||
 		(!strcmp(commandTag,"EXPLAIN")))
 	{
 		return true;
@@ -2623,7 +2820,7 @@ bool
 PGR_Did_Commit_Transaction(void)
 {
 
-	int sock = 0;
+	int sock = -1;
 	int cnt = 0;
 	ReplicateHeader header;
 	char * serverName = NULL;
@@ -2677,6 +2874,8 @@ PGR_Did_Commit_Transaction(void)
 	{
 		strncpy(header.from_host, PGRSelfHostName, HOSTNAME_MAX_LENGTH);
 	}
+	header.replicate_id = htonl(ReplicationLog_Info.PGR_Replicate_ID);
+	header.request_id = 0;
 
 	base = sp;
 	PGR_Sock_To_Replication_Server = sock;
@@ -2836,4 +3035,82 @@ PGR_Get_Cluster_Status(void)
 	return 0;
 }
 
+int
+PGR_Check_Replicate_Server_Status(ReplicateServerInfo * sp)
+{
+	ReplicateHeader header;
+	char * result = NULL;
+	int status;
+	int fdP;
+
+	result = malloc(PGR_MESSAGE_BUFSIZE + 4);
+	if (result == NULL)
+	{
+		if (Debug_pretty_print)
+			elog(DEBUG1,"malloc failed in ShowReplicationServerConfig()");
+		return STATUS_ERROR;
+	}
+
+	memset(&header, 0, sizeof(ReplicateHeader));
+	memset(result,  0, PGR_MESSAGE_BUFSIZE + 4);
+
+	header.cmdSys = CMD_SYS_PREREPLICATE;
+	header.cmdSts = CMD_STS_OTHER;
+	header.cmdType = CMD_TYPE_OTHER;
+	header.port = htons(PostPortNumber);
+	header.pid = htons(getpid());
+	header.query_size = 0;
+	strncpy(header.dbName ,(char *)(MyProcPort->database),sizeof(header.dbName));
+	strncpy(header.userName , (char *)(MyProcPort->user),sizeof(header.userName));
+	header.request_id = htonl(get_next_request_id());
+	header.rlog = 0;
+	if (PGRSelfHostName != NULL) {
+		strncpy(header.from_host, PGRSelfHostName, HOSTNAME_MAX_LENGTH);
+	}
+
+	/* open a new socket for lifecheck */
+	if ((status = PGR_Create_Socket_Connect(&fdP, sp->hostName, sp->portNumber)) == STATUS_ERROR) {
+		if (Debug_pretty_print) {
+			elog(DEBUG1,"create socket failed in ShowReplicationServerConfig()");
+		}
+		
+	/* status = STATUS_OK */
+	} else {
+		if ((status = send_replicate_packet(fdP, &header, (char *)NULL)) == STATUS_OK) {
+			/* receive result to check for possible deadlock */
+			status = (0 < recv_replicate_result(fdP, result ,0))
+				? STATUS_OK : STATUS_ERROR;
+		}
+	}
+
+	free(result);
+	PGR_Close_Sock(&fdP);
+
+	return status;
+}
+
+#if 0
+void
+PGR_Update_Replicate_Server_Status(ReplicateServerInfo * sp)
+{
+	ReplicateServerInfo * tmp;
+	ReplicateServerInfo * base;
+
+	tmp = PGR_get_replicate_server_info();
+	base = tmp;
+	get_new_replication_socket(base, tmp, PGR_QUERY_SOCKET);
+	
+/*
+	PGR_Set_Replication_Server_Status(sp, DATA_ERR);
+	search_new_replication_server(sp, PGR_QUERY_SOCKET);
+	if (sp == NULL) {
+		if (Debug_pretty_print) {
+			elog(DEBUG1,"all replication servers may be down");
+		}
+		PGR_Stand_Alone->is_stand_alone = true;
+	}
+*/
+}
+#endif
+
 #endif /* USE_REPLICATION */
