@@ -450,11 +450,6 @@ LockAcquire(LOCKMETHOD lockmethod, LOCKTAG *locktag,
 	int			myHolding[MAX_LOCKMODES];
 	int			i;
 
-#ifdef USE_REPLICATION
-	bool	pgr_dont_wait = false;
-Retry_LockAcquire:;
-
-#endif /* USE_REPLICATION */
 #ifdef LOCK_DEBUG
 	if (lockmethod == USER_LOCKMETHOD && Trace_userlocks)
 		elog(LOG, "LockAcquire: user lock [%u] %s",
@@ -650,30 +645,11 @@ Retry_LockAcquire:;
 		if ((PGR_Need_Notice == true) &&
 			(PGR_Check_Lock.check_lock_conflict == true))
 		{
-#ifdef CONTROL_LOCK_CONFLICT
-			if (PGR_Check_DeadLock( lockMethodTable, lockmode, lock, holder) == STATUS_OK)
-			{
-				PGR_Check_Lock.deadlock = DeadLockCheck(MyProc);
-			}
-			else
-			{
-				PGR_Check_Lock.deadlock = true;
-			}
-			if (PGR_Check_Lock.deadlock == true)
-			{
-				RemoveFromWaitQueue(MyProc);
-				MyProc->errType = STATUS_ERROR;
-				PGSemaphoreUnlock(&MyProc->sem);
-			}
-			pgr_dont_wait = dontWait;
-			dontWait = true;
-#else
-			if (PGR_Notice_Conflict() == STATUS_ERROR)
+			if (!PGR_Lock_Noticed && PGR_Notice_Conflict() == STATUS_ERROR)
 			{
 				return FALSE;
 			}
 			PGR_Lock_Noticed = true;
-#endif
 		}
 #endif /* USE_REPLICATION */
 
@@ -702,35 +678,6 @@ Retry_LockAcquire:;
 			Assert((lock->nRequested > 0) && (lock->requested[lockmode] >= 0));
 			Assert(lock->nGranted <= lock->nRequested);
 			LWLockRelease(masterLock);
-#ifdef USE_REPLICATION
-#ifdef CONTROL_LOCK_CONFLICT
-			if ((PGR_Need_Notice == true) &&
-				(PGR_Check_Lock.check_lock_conflict == true))
-			{
-				if (PGR_Check_Lock.deadlock != true)
-				{
-					enable_sig_alarm(DeadlockTimeout, false);
-				}
-				if (PGR_Notice_Conflict() == STATUS_ERROR)
-				{
-					return FALSE;
-				}
-				if (PGR_Check_Lock.deadlock != true)
-				{
-					disable_sig_alarm(false);
-				}
-				else
-				{
-					LOCK_PRINT("WaitOnLock: aborting on lock", lock, lockmode);
-					PGR_Need_Notice = false;
-					elog(ERROR, "deadlock detected");
-					return FALSE;
-				}
-				dontWait = pgr_dont_wait;
-				goto Retry_LockAcquire;
-			}
-#endif /* CONTROL_LOCK_CONFLICT */
-#endif /* USE_REPLICATION */
 			return FALSE;
 		}
 
