@@ -544,10 +544,10 @@ elog(int lev, const char *fmt,...)
 				}
 			}
 		}
-		if (PGR_Need_Notice == true)
-		{
+		//		if (PGR_Need_Notice == true)
+		//{
 			PGR_Notice_Transaction_Query_Done();
-		}
+			//}
 		if (PGR_Copy_Data_Need_Replicate)
 		{
 			PGR_Set_Copy_Data(PGRCopyData,NULL,0,1);
