@@ -80,6 +80,7 @@ int PGRset_status_to_child_tbl(pid_t pid, int status);
 int PGRadd_child_tbl(ClusterTbl * cluster_p, pid_t pid, int status);
 int PGRget_child_status(pid_t pid);
 void PGRreturn_connection_full_error(void);
+void PGRreturn_no_connection_error(void);
 void PGRquit_children_on_cluster(int rec_no);
 
 #ifdef NONE_BLOCK
@@ -93,6 +94,8 @@ static void cancel_request(CancelPacket *sp, int secondary_backend);
 static POOL_CONNECTION_POOL *connect_backend(PGR_StartupPacket *sp, POOL_CONNECTION *frontend);
 static int send_params(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend);
 static void child_end(int sig);
+static void PGRreturn_with_error(char *msg);
+
 
 /*--------------------------------------------------------------------
  * SYMBOL
@@ -207,7 +210,7 @@ retry_accept:
 	if (frontend == NULL)
 	{
 		/* accept() failed. return to the accept() loop */
-		PGRset_status_to_child_tbl(pid,TBL_ERROR);
+		PGRset_status_to_child_tbl(pid,TBL_FREE);
 		return STATUS_ERROR;
 	}
 
@@ -231,7 +234,7 @@ retry_startup:
 		/* failed to read the startup packet. return to the
 		   accept() loop */
 		pool_close(frontend);
-		PGRset_status_to_child_tbl(pid,TBL_ERROR);
+		PGRset_status_to_child_tbl(pid,TBL_FREE);
 		return STATUS_ERROR;
 	}
 	PGRset_status_to_child_tbl(pid,TBL_ACCEPT);
@@ -341,7 +344,7 @@ retry_startup:
 			case POOL_END:
 				/* do not cache connection to template0, template1, regression */
 				if (!strcmp(sp->database, "template0") || !strcmp(sp->database, "template1") ||
-					!strcmp(sp->database, "regression"))
+				    !strcmp(sp->database, "regression") || use_pool == NOT_USE_CONNECTION_POOL)
 				{
 					pool_close(frontend);
 					pool_send_frontend_exits(backend);
@@ -373,6 +376,7 @@ retry_startup:
 
 			/* fatal error occured. just exit myself... */
 			case POOL_FATAL:
+				show_error("%s:do_child: fatal error occured",func);
 				notice_backend_error();
 				break;
 
@@ -394,7 +398,7 @@ retry_startup:
 	if ((status == POOL_ERROR) || 
 		(status == POOL_FATAL))
 	{
-		PGRset_status_to_child_tbl(pid,TBL_ERROR);
+		PGRset_status_to_child_tbl(pid,TBL_FREE);
 		return STATUS_ERROR;
 	}
 	PGRset_status_to_child_tbl(pid,TBL_INIT);
@@ -517,7 +521,7 @@ void
 notice_backend_error(void)
 {
 	pid_t pid = getpid();
-	pid_t parent = getppid();
+
 	PGRset_status_to_child_tbl(pid,TBL_ERROR);
 	PGRset_status_on_cluster_tbl(TBL_ERROR_NOTICE,CurrentCluster);
 
@@ -637,10 +641,21 @@ PGRget_child_status(pid_t pid)
 
 void 
 PGRreturn_connection_full_error(void)
+{
+  PGRreturn_with_error( "Sorry, backend connection is full\n");
+}
+
+void 
+PGRreturn_no_connection_error(void) {
+  PGRreturn_with_error("pglb could not connect to server: no cluster available.\n");
+}
+
+static void 
+PGRreturn_with_error (char *msg) 
 {
 	PGR_StartupPacket *sp = NULL;
 	POOL_CONNECTION *frontend = NULL;
-	static char *msg = "Sorry, backend connection is full";
+
 
 	/* perform accept() */
 	frontend = do_accept(Frontend_FD.unix_fd,Frontend_FD.inet_fd);
@@ -886,6 +901,7 @@ static PGR_StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 	/* read startup packet length */
 	if (pool_read(cp, &len, sizeof(len)))
 	{
+		free(sp);
 		return NULL;
 	}
 	len = ntohl(len);
@@ -894,18 +910,22 @@ static PGR_StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 	if (len <= 0)
 	{
 		show_error("%s:read_startup_packet: incorrect packet length (%d)", func,len);
+		free(sp);
+		return NULL;
 	}
 
 	sp->startup_packet = calloc(len, 1);
 	if (!sp->startup_packet)
 	{
 		show_error("%s:read_startup_packet: out of memory",func);
+		free(sp);
 		return NULL;
 	}
 
 	/* read startup packet */
 	if (pool_read(cp, sp->startup_packet, len))
 	{
+		free(sp);
 		return NULL;
 	}
 
@@ -924,6 +944,7 @@ static PGR_StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 			if (!sp->database)
 			{
 				show_error("%s:read_startup_packet: out of memory",func);
+				free(sp);
 				return NULL;
 			}
 			strncpy(sp->database, sp2->database, SM_DATABASE);
@@ -932,6 +953,7 @@ static PGR_StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 			if (!sp->user)
 			{
 				show_error("%s:read_startup_packet: out of memory",func);
+				free(sp);
 				return NULL;
 			}
 			strncpy(sp->user, sp2->user, SM_USER);
@@ -950,6 +972,7 @@ static PGR_StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 					if (!sp->user)
 					{
 						show_error("%s:read_startup_packet: out of memory",func);
+						free(sp);
 						return NULL;
 					}
 				}
@@ -960,6 +983,7 @@ static PGR_StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 					if (!sp->database)
 					{
 						show_error("%s:read_startup_packet: out of memory",func);
+						free(sp);
 						return NULL;
 					}
 				}
@@ -973,18 +997,21 @@ static PGR_StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 			if (!sp->database)
 			{
 				show_error("%s:read_startup_packet: out of memory",func);
+				free(sp);
 				return NULL;
 			}
 			sp->user = calloc(1, 1);
 			if (!sp->user)
 			{
 				show_error("%s:read_startup_packet: out of memory",func);
+				free(sp);
 				return NULL;
 			}
 			break;
 
 		default:
 			show_error("%s:read_startup_packet: invalid major no: %d",func, sp->major);
+			free(sp);
 			return NULL;
 	}
 
