@@ -59,6 +59,7 @@
  * PROTOTYPE DECLARATION
  *--------------------------------------
  */
+int PGRis_cluster_alive(void) ;
 ClusterTbl * PGRscan_cluster(void);
 void PGRset_key_of_cluster(ClusterTbl * ptr, RecoveryPacket * packet);
 ClusterTbl * PGRadd_cluster_tbl (ClusterTbl * conf_data);
@@ -69,6 +70,24 @@ static int set_cluster_tbl(ClusterTbl * ptr , ClusterTbl * conf_data);
 static ClusterTbl * search_free_cluster_tbl(void );
 static void write_cluster_status_file(ClusterTbl * ptr);
 
+int PGRis_cluster_alive(void) 
+{
+	ClusterTbl * ptr = NULL;
+	int use=0;
+	ptr = Cluster_Tbl;
+
+	PGRsem_lock(ClusterSemid,MAX_DB_SERVER);
+	while (ptr->useFlag != TBL_END)
+	{
+		if (ptr->useFlag == TBL_USE)
+		{
+		  use++;
+		}
+		ptr++;
+	}
+	PGRsem_unlock(ClusterSemid,MAX_DB_SERVER);
+	return use==0 ? STATUS_ERROR : STATUS_OK;
+}
 
 ClusterTbl * 
 PGRscan_cluster(void)
@@ -91,7 +110,7 @@ PGRscan_cluster(void)
 	show_debug("%s:%d ClusterDB can be used",func,ClusterNum);
 #endif			
 	PGRsem_lock(ClusterSemid,MAX_DB_SERVER);
-	while ((cnt < ClusterNum) && (ptr->useFlag != TBL_END))
+	while ((cnt <= ClusterNum) && (ptr->useFlag != TBL_END))
 	{
 #ifdef PRINT_DEBUG
 		show_debug("%s:%s [%d],useFlag->%d max->%d use_num->%d\n",
@@ -208,9 +227,10 @@ PGRset_status_on_cluster_tbl (int status, ClusterTbl * ptr)
 			write_cluster_status_file(ptr);
 			if (status == TBL_USE)
 			{
-				ClusterNum ++ ;
+				if (ClusterNum < Max_DB_Server)
+					ClusterNum ++ ;
 			}
-			else
+			else if (status != TBL_STOP)
 			{
 				if (ClusterNum > 0)
 					ClusterNum -- ;
@@ -263,14 +283,17 @@ PGRsearch_cluster_tbl(ClusterTbl * conf_data)
 	ptr = Cluster_Tbl;
 	while ((cnt <= ClusterNum) && (rec_num < Max_DB_Server))
 	{
-		if ((!strcmp(ptr->hostName,conf_data->hostName)) &&
-			(ptr->port == conf_data->port))
+		if (ptr->port > 0)
 		{
-			return ptr;
-		}
-		if (ptr->useFlag == TBL_USE)
-		{
-			cnt ++;
+			if ((!strcmp(ptr->hostName,conf_data->hostName)) &&
+				(ptr->port == conf_data->port))
+			{
+				return ptr;
+			}
+			if (ptr->useFlag == TBL_USE)
+			{
+				cnt ++;
+			}
 		}
 		ptr ++;
 		rec_num ++;
