@@ -125,7 +125,6 @@ PGRload_balance(void)
 		PGRdo_child(NOT_USE_CONNECTION_POOL );
 		PGRrelease_connection(cluster_p);
 		PGRset_status_to_child_tbl(getpid(), TBL_FREE);
-		pool_finish();
 		exit(0);
 	}
 	else if (pid > 0)
@@ -179,7 +178,7 @@ PGRload_balance_with_pool(void)
 		if ( count > PGLB_CONNECT_RETRY_TIME)
 		{
 			show_error("%s:no cluster available",func);
-			PGRreturn_connection_full_error();
+       			PGRreturn_no_connection_error();
 			return STATUS_ERROR;
 		}
 		cluster_p = PGRscan_cluster();
