@@ -8,7 +8,7 @@
 #ifndef PGLB_H
 #define PGLB_H
 
-#define PGLB_VERSION	"1.0.9"
+#define PGLB_VERSION	"1.0.10a"
 
 #include "../libpgc/libpgc.h"
 
@@ -370,9 +370,11 @@ extern int PGRset_status_to_child_tbl(pid_t pid, int status);
 extern int PGRadd_child_tbl(ClusterTbl * cluster_p, pid_t pid, int status);
 extern int PGRget_child_status(pid_t pid);
 extern void PGRreturn_connection_full_error(void);
+extern void PGRreturn_no_connection_error(void);
 extern void PGRquit_children_on_cluster(int rec_no);
 
 /* extern of cluster_table.c */
+extern int PGRis_cluster_alive(void) ;
 extern ClusterTbl * PGRscan_cluster(void);
 extern void PGRset_key_of_cluster(ClusterTbl * ptr, RecoveryPacket * packet);
 extern ClusterTbl * PGRadd_cluster_tbl (ClusterTbl * conf_data);
