@@ -25,18 +25,15 @@
 #include <errno.h>
 #include <sys/file.h>
 
+
+
 #include "libpq-fe.h"
+#include "libpq-int.h"
+#include "fe-auth.h"
+
 #include "replicate_com.h"
 #include "pgreplicate.h"
 
-#ifndef HAVE_UNION_SEMUN
-union semun {
-	int val;
-	struct semid_ds *buf;
-	unsigned short int *array;
-	struct seminfo *__buf;
-};
-#endif
 int PGRset_Conf_Data(char * path);
 
 /*--------------------------------------------------------------------
@@ -62,7 +59,7 @@ PGRset_Conf_Data(char * path)
 	int cascade_cnt = 0;
 	int rec_no = 0;
 	int lb_rec_no = 0;
-	int cascade_rec_no = 0;
+	int cascade_rec_no = -1;
 	int i = 0;
 	int size = 0;
 	char fname[256];
@@ -85,8 +82,9 @@ PGRset_Conf_Data(char * path)
 			show_error("%s:fopen failed: (%s)",func,strerror(errno));
 			return STATUS_ERROR;
 		}
-#ifdef PRINT_DEBUG
 		show_debug("%s open ok\n",fname);
+#ifdef PRINT_DEBUG
+
 #endif			
 	}
 
@@ -97,9 +95,6 @@ PGRset_Conf_Data(char * path)
 		show_error("%s:fopen failed: (%s)",func,strerror(errno));
 		return STATUS_ERROR;
 	}
-#ifdef PRINT_DEBUG
-	show_debug("%s open ok\n",fname);
-#endif			
 
 	/*
 	 * read configuration file
@@ -222,6 +217,20 @@ PGRset_Conf_Data(char * path)
 		semctl(SemID, i, SETVAL, sem_arg);
 	}
 
+	/* create semapho */
+	if ((CascadeSemID = semget(IPC_PRIVATE,2,IPC_CREAT | IPC_EXCL | 0600)) < 0)
+	{
+		show_error("%s:semget() failed. (%s)",func,strerror(errno));
+		return STATUS_ERROR;
+	}
+	for ( i = 0 ; i < 2 ; i ++)
+	{
+		semctl(CascadeSemID, i, GETVAL, sem_arg);
+		sem_arg.val = 1;
+		semctl(CascadeSemID, i, SETVAL, sem_arg);
+	}
+
+
 	if ((VacuumSemID = semget(IPC_PRIVATE,2,IPC_CREAT | IPC_EXCL | 0600)) < 0)
 	{
 		show_error("%s:semget() failed. (%s)",func,strerror(errno));
@@ -233,12 +242,27 @@ PGRset_Conf_Data(char * path)
 		sem_arg.val = 1;
 		semctl(VacuumSemID, i, SETVAL, sem_arg);
 	}
+	size = sizeof(ReplicationLogInf);
+	Replicateion_Log = malloc(size);
+	if (Replicateion_Log == NULL)
+	{
+		show_error("%s:malloc failed: (%s)",func,strerror(errno));
+		return STATUS_ERROR;
+	}
+	memset(Replicateion_Log , 0 , size );
+	Replicateion_Log->RLog_Sock_Path = NULL;
+#ifdef PRINT_DEBUG
+	show_debug("%s:RLog Memory Allocation ok",func);
+#endif			
+
+
 	/*
 	 * set each datas into the tables
 	 */
 	conf = ConfData_Top;
 	while (conf != (ConfDataType *)NULL) 
 	{
+		show_debug("registering (key,value)=(%s,%s)",conf->key,conf->value);
 		/* get cluster db data */
 		if (!strcmp(conf->table,CLUSTER_SERVER_TAG))
 		{
@@ -253,7 +277,19 @@ PGRset_Conf_Data(char * path)
 			}
 			if (!strcmp(conf->key,HOST_NAME_TAG))
 			{
+			        int ip;
 				memcpy(host_tbl[rec_no].hostName,conf->value,sizeof(host_tbl[rec_no].hostName));
+				show_debug("registering hostname %s",host_tbl[rec_no].hostName);
+				ip=PGRget_ip_by_name(conf->value);
+
+				sprintf(host_tbl[rec_no].resolvedName,
+					 "%d.%d.%d.%d",
+					 (ip      ) & 0xff ,
+					 (ip >>  8) & 0xff ,
+					 (ip >> 16) & 0xff ,
+					 (ip >> 24) & 0xff );
+				show_debug("resolved name is %s",host_tbl[rec_no].resolvedName);
+
 				conf = (ConfDataType*)conf->next;
 				continue;
 			}
@@ -290,22 +326,36 @@ PGRset_Conf_Data(char * path)
 			}
 			if (!strcmp(conf->key,PORT_TAG))
 			{
-				(Cascade_Tbl+cascade_rec_no)->portNumber = atoi(conf->value);
-				conf = (ConfDataType*)conf->next;
-				if (cascade_rec_no == 0)
+				if (atoi(conf->value) > 0)
 				{
-					PGRset_cascade_server_status(Cascade_Tbl+cascade_rec_no,DB_TBL_TOP);
+					(Cascade_Tbl+cascade_rec_no)->portNumber = atoi(conf->value);
 				}
 				else
 				{
-					PGRset_cascade_server_status(Cascade_Tbl+cascade_rec_no,DB_TBL_USE);
+					(Cascade_Tbl+cascade_rec_no)->portNumber = DEFAULT_PGRP_PORT;
+				}
+				(Cascade_Tbl+cascade_rec_no)->sock = -1;
+
+				conf = (ConfDataType*)conf->next;
+				PGRset_cascade_server_status(Cascade_Tbl+cascade_rec_no,DB_TBL_USE);
+				if (cascade_rec_no == 0)
+				{
+					Cascade_Inf->top = Cascade_Tbl;
 				}
-				(Cascade_Tbl+cascade_rec_no +1)->useFlag = DB_TBL_END;
 				continue;
 			}
 			if (!strcmp(conf->key,RECOVERY_PORT_TAG))
 			{
-				(Cascade_Tbl+cascade_rec_no)->recoveryPortNumber = atoi(conf->value);
+				if (atoi(conf->value) > 0)
+				{
+					(Cascade_Tbl+cascade_rec_no)->recoveryPortNumber = atoi(conf->value);
+				}
+				else
+				{
+					(Cascade_Tbl+cascade_rec_no)->recoveryPortNumber = DEFAULT_PGRP_RECOVERY_PORT;
+				}
+				(Cascade_Tbl+cascade_rec_no)->rlog_sock=-1;
+				(Cascade_Tbl+cascade_rec_no +1)->useFlag = DB_TBL_END;
 				conf = (ConfDataType*)conf->next;
 				continue;
 			}
@@ -348,10 +398,57 @@ PGRset_Conf_Data(char * path)
 			/* get port number for recovery cluster db server */
 			if (!strcmp(conf->key,RECOVERY_PORT_TAG))
 			{
-				Recovery_Port_Number = atoi(conf->value);
+				if (atoi(conf->value) > 0)
+				{
+					Recovery_Port_Number = atoi(conf->value);
+				}
+				else
+				{
+					Recovery_Port_Number =DEFAULT_PGRP_RECOVERY_PORT;
+				}
 				conf = (ConfDataType*)conf->next;
 				continue;
 			}
+			if (!STRCMP(conf->key,LIFECHECK_PORT_TAG))
+			{
+				if (atoi(conf->value) > 0)
+				{
+					LifeCheck_Port_Number = atoi(conf->value);
+				}
+				else
+				{
+					LifeCheck_Port_Number = DEFAULT_PGRP_LIFECHECK_PORT;
+				}
+				conf = (ConfDataType*)conf->next;
+				continue;
+			}
+			if (!STRCMP(conf->key,RLOG_PORT_TAG))
+			{
+				if (atoi(conf->value) > 0)
+				{
+					Replicateion_Log->RLog_Port_Number = atoi(conf->value);
+				}
+				else
+				{
+					Replicateion_Log->RLog_Port_Number = DEFAULT_PGRP_RLOG_PORT;
+				}
+				conf = (ConfDataType*)conf->next;
+				continue;
+			}
+			/*
+			if (!STRCMP(conf->key,STATUS_LOG_FILE_TAG))
+			{
+				Com_Info.Log_Info.PGRStatusFileName = strdup(conf->value);
+				conf = (ConfDataType*)conf->next;
+				continue;
+			}
+			if (!STRCMP(conf->key,ERROR_LOG_FILE_TAG))
+			{
+				Com_Info.Log_Info.PGRLogFileName = strdup(conf->value);
+				conf = (ConfDataType*)conf->next;
+				continue;
+			}
+			*/
 			/* get response mode */
 			if (!strcmp(conf->key,RESPONSE_MODE_TAG))
 			{
@@ -384,9 +481,29 @@ PGRset_Conf_Data(char * path)
 		}
 		conf = (ConfDataType*)conf->next;
 	}
-#ifdef PRINT_DEBUG
-	show_debug("Conf data read ok");
-#endif			
+
+	/*
+	 * open log file
+	 */
+	if (Log_Print)
+	{
+		snprintf(fname,sizeof(fname),"%s/%s",path,PGREPLICATE_LOG_FILE);
+
+		LogFp = fopen(fname,"a");
+		if (LogFp == NULL)
+		{
+			show_error("%s:fopen failed: (%s)",func,strerror(errno));
+			return STATUS_ERROR;
+		}
+	}
+
+        snprintf(fname,sizeof(fname),"%s/%s",path,PGREPLICATE_STATUS_FILE);
+	StatusFp = fopen(fname,"a");
+	if (StatusFp == NULL)
+	{
+		show_error("%s:fopen failed: (%s)",func,strerror(errno));
+		return STATUS_ERROR;
+	}
 
 	/* create cluster db server table */
 	Host_Tbl_Begin = (HostTbl *)NULL;
@@ -446,7 +563,7 @@ PGRset_Conf_Data(char * path)
 
 	/* allocate send query id */
 	size = sizeof(unsigned int) * (MAX_DB_SERVER +1);
-	PGR_Send_Query_ID = malloc (sizeof(unsigned int) * MAX_DB_SERVER);
+	PGR_Send_Query_ID = malloc (size);
 	if (PGR_Send_Query_ID == NULL)
 	{
 		show_error("%s:malloc() failed. reason: %s", func, strerror(errno));
@@ -460,26 +577,31 @@ PGRset_Conf_Data(char * path)
 
 	/* set self data into cascade table */
 
-	if (cascade_rec_no > 0)
-	{
-		cascade_rec_no ++;
-	}
+	cascade_rec_no ++;
 	gethostname((Cascade_Tbl+cascade_rec_no)->hostName,sizeof(Cascade_Tbl->hostName));
 	(Cascade_Tbl+cascade_rec_no)->portNumber = Port_Number;
 	(Cascade_Tbl+cascade_rec_no)->recoveryPortNumber = Recovery_Port_Number;
+	(Cascade_Tbl+cascade_rec_no)->sock = -1;
 
 	PGRset_cascade_server_status(Cascade_Tbl+cascade_rec_no,DB_TBL_USE);
-	if (cascade_rec_no == 0)
+	/* terminate */
+	(Cascade_Tbl+(cascade_rec_no+1))->useFlag = DB_TBL_END;
+
+	Cascade_Inf->top = Cascade_Tbl;
+	Cascade_Inf->end = Cascade_Tbl+cascade_rec_no;
+      	Cascade_Inf->upper = NULL;
+	Cascade_Inf->lower = NULL;
+	if (cascade_rec_no >= 1)
 	{
-		PGRset_cascade_server_status(Cascade_Tbl,DB_TBL_TOP);
+		Cascade_Inf->upper = (Cascade_Tbl+cascade_rec_no - 1);
 	}
 	(Cascade_Tbl+(cascade_rec_no+1))->useFlag = DB_TBL_END;
 
 	Cascade_Inf->myself = (Cascade_Tbl+cascade_rec_no);
-	Cascade_Inf->upper = NULL;
-	Cascade_Inf->lower = NULL;
 	Cascade_Inf->useFlag = DB_TBL_USE;
 
+	PGR_Response_Inf->response_mode = PGR_NORMAL_MODE;
+
 	return STATUS_OK;
 }
 
