@@ -31,6 +31,7 @@
 #include <sys/param.h>
 #include <arpa/inet.h>
 #include <sys/file.h>
+#include <pthread.h>
 
 #ifdef HAVE_NETINET_TCP_H
 #include <netinet/tcp.h>
@@ -45,10 +46,13 @@
 
 #include "miscadmin.h"
 #include "nodes/nodes.h"
+
 #include "libpq-fe.h"
 #include "libpq/libpq-fs.h"
 #include "libpq-int.h"
 #include "fe-auth.h"
+
+
 #include "access/xact.h"
 #include "replicate_com.h"
 #include "pgreplicate.h"
@@ -69,31 +73,50 @@
  * GLOBAL VARIABLE DECLARATION
  *--------------------------------------
  */
+/* for replicate_com.h */
+
+ConfDataType * ConfData_Top = (ConfDataType *)NULL;
+ConfDataType * ConfData_End = (ConfDataType *)NULL;
+
+/* replication server data */
+uint16_t Port_Number = 0;
+uint16_t LifeCheck_Port_Number = 0;
+uint16_t Recovery_Port_Number = 0;
+
+/* global table data */
 HostTbl *Host_Tbl_Begin = NULL;
-TransactionTbl * Transaction_Tbl_Begin = NULL;
+Dllist * Transaction_Tbl_Begin = NULL;
 TransactionTbl * Transaction_Tbl_End = NULL;
 RecoveryTbl * LoadBalanceTbl = NULL;
 RecoveryStatusInf * Recovery_Status_Inf = NULL;
 ReplicateHeader * PGR_Log_Header = NULL;
 ReplicateServerInfo * Cascade_Tbl = NULL;;
 CommitLogInf * Commit_Log_Tbl = NULL;
-ReplicationLogInf * Replicateion_Log = NULL;
-int Port_Number = 0;
+QueryLogType * Query_Log_Top = NULL;
+QueryLogType * Query_Log_End = NULL;
 CascadeInf * Cascade_Inf = NULL;
-int Recovery_Port_Number = 0;
+ReplicationLogInf * Replicateion_Log = NULL;
+/* IPC's id data */
 int RecoveryShmid = 0;
-int RecoveryMsgid = 0;
+int ReplicateSerializationShmid=0;
+int RecoveryMsgShmid = 0;
+int *RecoveryMsgid = NULL;
 int HostTblShmid = 0;
+int LockWaitTblShmid = 0;
+int LoadBalanceTblShmid = 0;
 int CascadeTblShmid = 0;
 int CascadeInfShmid = 0;
 int CommitLogShmid = 0;
-ConfDataType * ConfData_Top = (ConfDataType *)NULL;
-ConfDataType * ConfData_End = (ConfDataType *)NULL;
+int QueryLogMsgid = 0;
+int QueryLogAnsMsgid = 0;
+int PGconnMsgid = 0;
 int MaxBackends = 0;
 char * PGR_Result = NULL;
 int SemID = 0;
-int RecoverySemID = 0;
+int RecoverySemID= 0;
+int RecovErysemid = 0;
 int VacuumSemID = 0;
+int CascadeSemID= 0;
 char * PGR_Data_Path = NULL;
 char * PGR_Write_Path = NULL;
 int IS_SESSION_AUTHORIZATION = 0;
@@ -104,6 +127,15 @@ bool PGR_Use_Replication_Log = false;
 bool	PGR_AutoCommit = true;
 unsigned int * PGR_Send_Query_ID = NULL;
 unsigned int PGR_Query_ID = 0;
+bool exit_processing = false;
+int pgreplicate_pid = 0;
+
+int ReplicateSock = -1;
+
+RecoveryQueueInf RecoveryQueue;
+char * Backend_Socket_Dir = NULL;
+
+unsigned int * PGR_ReplicateSerializationID = NULL;
 
 int Log_Print = 0;
 int Debug_Print = 0;
@@ -113,6 +145,12 @@ FILE * StatusFp = (FILE *)NULL;
 extern char *optarg;
 char * PGRuserName = NULL;
 
+int fork_wait_time = 0;
+int Idle_Flag = IDLE_MODE;
+bool Exit_Request = false;
+
+pthread_mutex_t transaction_table_mutex;
+
 /*--------------------------------------
  * PROTOTYPE DECLARATION
  *--------------------------------------
@@ -125,7 +163,6 @@ static void daemonize(void);
 static void write_pid_file(void);
 static void stop_pgreplicate(void);
 static bool is_exist_pid_file(void);
-static void child_wait(SIGNAL_ARGS);
 static void usage(void);
 
 /*--------------------------------------------------------------------
@@ -149,9 +186,20 @@ replicate_loop(int fd)
 	int sock = -1;
 	int rtn = 0;
 	int cnt = 0;
+	int result;
+	bool exist_sys_log=false;
+	bool exist_replicate=false;
+
+
+	result = PGR_Create_Acception(fd,&sock,"",Port_Number);
+	if (result == STATUS_ERROR)
+	{
+		show_error("%s: accept failed (%s)", func, strerror(errno));
+		if (sock != -1)
+			close(sock);
+		return 1;
+	}
 
-	signal(SIGCHLD, SIG_IGN);	
-	PGR_Create_Acception(fd,&sock,"",Port_Number);
 	pgid = getpgid(0);
 	pid = fork();
 	if (pid <0)
@@ -164,15 +212,34 @@ replicate_loop(int fd)
 		int status = LOOP_CONTINUE;
 		bool PGR_Cascade = false;
 		ReplicateHeader  header;
+		ReplicateHeader  header_save_for_recovering;
 		char * query = NULL;
 
+		if (fork_wait_time > 0) {
+#ifdef PRINT_DEBUG
+			show_debug("wait fork(): pid = %d", getpid());
+#endif			
+			sleep(fork_wait_time);
+		}
+
+		close(fd);
+
 		signal(SIGHUP, quick_exit);	
 		signal(SIGINT, quick_exit);	
 		signal(SIGQUIT, quick_exit);	
 		signal(SIGTERM, quick_exit);	
 		signal(SIGALRM, quick_exit); 
-		signal(SIGPIPE, quick_exit); 
+		signal(SIGPIPE, SIG_IGN); 
 		setpgid(0,pgid);
+		
+		if (PGRinit_transaction_table() != STATUS_OK)
+		{
+			show_error("transaction table memory allocate failed");
+			PGR_Close_Sock(&sock);
+			exit(1);
+		}
+
+		pthread_mutex_init(&transaction_table_mutex, NULL);
 
 		/* child loop */
 		for (;;)
@@ -194,64 +261,147 @@ replicate_loop(int fd)
 			FD_ZERO(&rmask);
 			FD_SET(sock,&rmask);
 			rtn = select(sock+1, &rmask, (fd_set *)NULL, (fd_set *)NULL, &timeout);
+			if (rtn < 0)
+			{
+				if (errno == EINTR)
+					continue;
+			}
+
 			if (rtn && FD_ISSET(sock, &rmask))
 			{
 				query = NULL;
-				query = PGRread_packet(sock,&header);
+				query = PGRread_packet(sock,&header);				
 				if ((query == NULL) || (header.cmdSts == 0))
 				{
-					cnt ++;
-					if (cnt < 5)
+
+					if (exist_sys_log)
 					{
-						/*
-						PGR_Close_Sock(&sock);
-						PGR_Create_Acception(fd,&sock,"",Port_Number);
-						*/
-						continue;
+						show_error("%s:upper cascade closed? , errno=%d(%s)",func,errno,strerror(errno));
+						memset(&header, 0, sizeof(ReplicateHeader));
+						header.cmdSys = CMD_SYS_CALL;
+						header.cmdSts = CMD_STS_QUERY_SUSPEND;
+						header.query_size = htonl(0);
+						PGRsend_rlog_to_local(&header, NULL);
+						exist_sys_log = false;
 					}
 					else
 					{
-						break;
+						int recovery_status = PGRget_recovery_status();
+#ifdef PRINT_DEBUG
+						show_debug("%s:session closed",func);
+#endif
+						if (recovery_status == RECOVERY_START)
+						{
+							header_save_for_recovering.cmdSts=CMD_TYPE_OTHER;
+							header_save_for_recovering.cmdType=CMD_TYPE_CONNECTION_CLOSE;
+							header_save_for_recovering.query_size = htonl(21);
+							PGRdo_replicate(sock,&header_save_for_recovering,"PGR_CLOSE_CONNECTION");
+						}
+
+						if(exist_replicate) 
+						{
+							int recovery_status;
+
+							PGRsem_lock(RecoverySemID,SEM_NUM_OF_RECOVERY);
+							Recovery_Status_Inf->transaction_count --;
+							PGRsem_unlock(RecoverySemID, SEM_NUM_OF_RECOVERY);										
+							show_debug("transaction_status: %d",Recovery_Status_Inf->transaction_count);
+							recovery_status=PGRget_recovery_status();
+							
+							header_save_for_recovering.cmdSts=CMD_TYPE_OTHER;
+							header_save_for_recovering.cmdType=CMD_TYPE_CONNECTION_CLOSE;
+							PGRdo_replicate(sock,&header_save_for_recovering,"PGR_CLOSE_CONNECTION");
+						}
+						PGRsend_notice_quit();
 					}
-					/*
-					usleep(PGR_RECV_WAIT_MSEC);
-					continue;
 					break;
-					*/
 				}
 				cnt = 0;
+				memcpy(&header_save_for_recovering,
+					   &header,
+					   sizeof(ReplicateHeader));
 				switch (header.cmdSys)
 				{
-					case CMD_SYS_REPLICATE:
-						status = PGRdo_replicate(sock,&header,query);
-						break;
-					case CMD_SYS_LOG:
-						PGRwrite_rlog(&header, query);
-						break;
-					case  CMD_SYS_CASCADE:
-						PGR_Cascade = true;
-						PGRcascade_main(sock,&header,query);
-						break;
-					case  CMD_SYS_CALL:
-						if (header.cmdSts == CMD_STS_TRANSACTION_ABORT)
-						{
-							PGRreconfirm_commit(sock,&header);
-						}
-						else if (header.cmdSts == CMD_STS_NOTICE)
-						{
-							if (header.cmdType == CMD_TYPE_DEADLOCK_DETECT)
-							{
-								status = LOOP_END;
-							}
-						}
-						else if (header.cmdSts == CMD_STS_RESPONSE)
-						{
-							if (header.cmdType == CMD_TYPE_FRONTEND_CLOSED)
-							{
-								status = LOOP_END;
-							}
-						}
-						break;
+				        case CMD_SYS_PREREPLICATE:
+								if(Cascade_Inf!=NULL ||
+								   Cascade_Inf->upper == NULL) 
+								{
+										/* 1 means "I am primary replicate server." */
+										PGRreturn_result(sock,"1", PGR_NOWAIT_ANSWER);
+								}
+								else
+								{
+										/* 1 means "I am not primary replicate server." */
+										PGRreturn_result(sock,"0", PGR_NOWAIT_ANSWER);
+				                }
+								break;
+						case CMD_SYS_REPLICATE:
+								status = PGRdo_replicate(sock,&header,query);
+								if (header.cmdSts == CMD_STS_TRANSACTION )
+								{
+										switch(header.cmdType)
+										{
+												case CMD_TYPE_BEGIN:
+														exist_replicate=true;
+														memcpy(&header_save_for_recovering,
+															   &header,
+															   sizeof(ReplicateHeader));
+														break;
+												case CMD_TYPE_COMMIT:
+												case CMD_TYPE_ROLLBACK:
+														exist_replicate=false;
+														break;
+										}
+								}
+								else
+								{
+										if ( header.cmdType == CMD_TYPE_COPY )
+										{
+														exist_replicate=true;
+														memcpy(&header_save_for_recovering,
+															   &header,
+															   sizeof(ReplicateHeader));
+										}
+										else if (header.cmdType == CMD_TYPE_COPY_DATA_END) 
+										{
+														exist_replicate=false;
+										}
+								}
+
+								break;
+						case CMD_SYS_LOG:
+								exist_sys_log = true;
+								PGRsend_rlog_to_local(&header, query);
+								/* set own replicate id by rlog */
+								PGRset_replication_id(ntohl(header.replicate_id));
+								PGRsend_notice_rlog_done(sock);
+								break;
+						case  CMD_SYS_CASCADE:
+								PGR_Cascade = true;
+								PGRcascade_main(sock,&header,query);
+								break;
+						case  CMD_SYS_CALL:
+								if (header.cmdSts == CMD_STS_TRANSACTION_ABORT)
+								{
+										PGRreconfirm_commit(sock,&header);
+								
+								}
+								else if (header.cmdSts == CMD_STS_NOTICE)
+								{
+
+								}
+								else if (header.cmdSts == CMD_STS_RESPONSE)
+								{
+										if (header.cmdType == CMD_TYPE_FRONTEND_CLOSED)
+										{
+												show_debug("replicate_loop ends due to frontend_closed.");
+												PGRsend_notice_rlog_done(sock);
+												status = LOOP_END;
+										}
+								}
+								break;
+						default:
+								show_debug("WARNING: unknown Header->cmdSys %c",header.cmdSys);
 				}
 			}
 			if (status == LOOP_END)
@@ -259,6 +409,8 @@ replicate_loop(int fd)
 				break;
 			}
 		}
+
+		pthread_mutex_destroy(&transaction_table_mutex);
 		PGR_Close_Sock(&sock);
 		if (query != NULL)
 		{
@@ -314,7 +466,7 @@ startup_replication_server(void)
 	memcpy(header.from_host,hostName,sizeof(header.from_host));
 	memcpy(header.userName,userName,sizeof(header.userName));
 	strcpy(header.dbName,"template1");
-	PGRreplicate_packet_send( &header, query,-1);
+	replicate_packet_send_internal( &header, query,-1,PGRget_recovery_status(),true);
 }
 
 /*--------------------------------------------------------------------
@@ -335,23 +487,28 @@ replicate_main(void)
 	char * func = "replicate_main()";
 #endif			
 	int status;
-	int fd = -1;
 	int rtn;
+	show_debug ("%s:entering replicate_main",func);
+
+	/* cascade start up notice */
+	if (Cascade_Inf->upper != NULL)
+	{
+		show_debug("initialize cascade information");
+		PGRstartup_cascade();
+	}
+
+	status = PGR_Create_Socket_Bind(&ReplicateSock, "", Port_Number);
 
-	status = PGR_Create_Socket_Bind(&fd, "", Port_Number);
 	if (status != STATUS_OK)
 	{
+		show_debug("%s %d port bind failed. quit.",func,Port_Number);
+		stop_pgreplicate();
 		PGRreplicate_exit(0);
 	}
 #ifdef PRINT_DEBUG
 	show_debug("%s %d port bind OK",func,Port_Number);
 #endif			
 	
-	/* cascade start up notice */
-	if (Cascade_Inf->upper != NULL)
-	{
-		PGRstartup_cascade();
-	}
 
 	/* replication start up notice */
 	startup_replication_server();
@@ -361,6 +518,9 @@ replicate_main(void)
 		fd_set	  rmask;
 		struct timeval timeout;
 
+		if (exit_processing == true)
+			PGRreplicate_exit(0);
+
 		timeout.tv_sec = PGR_RECV_TIMEOUT;
 		timeout.tv_usec = 0;
 
@@ -369,18 +529,25 @@ replicate_main(void)
 		 * Wait for something to happen.
 		 */
 		FD_ZERO(&rmask);
-		FD_SET(fd,&rmask);
-		rtn = select(fd+1, &rmask, (fd_set *)NULL, (fd_set *)NULL, &timeout);
-		if (rtn && FD_ISSET(fd, &rmask))
+		FD_SET(ReplicateSock,&rmask);
+		rtn = select(ReplicateSock+1, &rmask, (fd_set *)NULL, (fd_set *)NULL, &timeout);
+		if (rtn < 0)
+			continue;
+
+		if (rtn && FD_ISSET(ReplicateSock, &rmask))
 		{
 			/*
 			 * get recovery status.
 			 */
 			PGRcheck_recovered_host();
+
+			if (exit_processing == true)
+				break;
+
 			/*
 			 * call replication module
 			 */
-			replicate_loop(fd);
+			replicate_loop(ReplicateSock);
 		}
 	}
 }
@@ -399,6 +566,9 @@ replicate_main(void)
 static void
 quick_exit(SIGNAL_ARGS)
 {
+#ifdef PRINT_DEBUG
+	show_debug("quick_exit:signo = %d", postgres_signal_arg);
+#endif
 	exit(0);
 }
 
@@ -567,7 +737,7 @@ is_exist_pid_file(void)
  *    none
  *--------------------------------------------------------------------
  */
-static void
+void
 child_wait(SIGNAL_ARGS)
 {
 	pid_t pid = 0;
@@ -630,11 +800,12 @@ main(int argc, char * argv[])
 	char * r_path = NULL;
 	char * w_path = NULL;
 	bool detach = true;
+	pid_t rlog_pid;
 
 	r_path = getenv("PGDATA");
 	if (r_path == NULL)
 		r_path = ".";
-	while ((opt = getopt(argc, argv, "U:D:W:lvnh")) != -1)
+	while ((opt = getopt(argc, argv, "U:D:W:w:lvnh")) != -1)
 	{
 		switch (opt)
 		{
@@ -644,7 +815,7 @@ main(int argc, char * argv[])
 					usage();
 					exit(1);
 				}
-				PGRuserName = optarg;
+				PGRuserName = strdup(optarg);
 				break;
 			case 'D':
 				if (!optarg)
@@ -662,6 +833,9 @@ main(int argc, char * argv[])
 				}
 				w_path = optarg;
 				break;
+			case 'w':
+				fork_wait_time = atoi(optarg);
+				break;
 			case 'l':
 				Log_Print = 1;
 				break;
@@ -714,6 +888,13 @@ main(int argc, char * argv[])
 		daemonize();
 	}
 	write_pid_file();
+	pgreplicate_pid = getpid();
+
+	signal(SIGINT, set_exit_processing);
+	signal(SIGQUIT, set_exit_processing);
+	signal(SIGTERM, set_exit_processing);
+	signal(SIGCHLD, child_wait);
+	signal(SIGPIPE, SIG_IGN);
 
 	if (PGRset_Conf_Data(PGR_Data_Path) != STATUS_OK)
 	{
@@ -725,7 +906,21 @@ main(int argc, char * argv[])
 		show_error("%s:PGRinit_recovery error",func);
 		PGRreplicate_exit(0);
 	}
+
 	pgid = getpgid(0);
+	if ( PGR_Use_Replication_Log == true )
+	{
+#ifdef PRINT_DEBUG
+		show_debug("Use Replication Log. Start PGR_RLog_Main()");
+#endif
+		rlog_pid = PGR_RLog_Main();
+		if (rlog_pid < 0)
+		{
+			show_error("%s:PGR_RLog_Main failed",func);
+			PGRreplicate_exit(0);
+		}
+	}
+
 	pid = fork();
 	if (pid == (pid_t) -1)
 	{
@@ -734,6 +929,7 @@ main(int argc, char * argv[])
 	}
 	if (pid == 0)
 	{
+		signal(SIGCHLD, SIG_DFL);
 		signal(SIGHUP, quick_exit);	
 		signal(SIGINT, quick_exit);	
 		signal(SIGQUIT, quick_exit);	
@@ -744,21 +940,34 @@ main(int argc, char * argv[])
 		 * call recovery module
 		 */
 		setpgid(0,pgid);
+
+		if (fork_wait_time > 0) {
+#ifdef PRINT_DEBUG
+			show_debug("recovery process: wait fork(): pid = %d", getpid());
+#endif			
+			sleep(fork_wait_time);
+		}
+
 		PGRrecovery_main();
 	}
 	else
 	{
-		signal(SIGINT, PGRreplicate_exit);
-		signal(SIGQUIT, PGRreplicate_exit);
-		signal(SIGTERM, PGRreplicate_exit);
-		signal(SIGCHLD, child_wait);
-		signal(SIGPIPE, SIG_IGN);	
 		/*
 		 * call replicate module
 		 */
+		Replicateion_Log->r_log_sock =-1;
+
+		if (fork_wait_time > 0) {
+#ifdef PRINT_DEBUG
+			show_debug("replicate process: wait fork(): pid = %d", getpid());
+#endif			
+			sleep(fork_wait_time);
+		}
+
 		replicate_main();
 	}
 
 	PGRreplicate_exit(0);
 	return STATUS_OK;
 }
+
