@@ -40,10 +40,12 @@
 
 #include "miscadmin.h"
 #include "nodes/nodes.h"
+
 #include "libpq-fe.h"
 #include "libpq/libpq-fs.h"
 #include "libpq-int.h"
 #include "fe-auth.h"
+
 #include "access/xact.h"
 #include "replicate_com.h"
 #include "pgreplicate.h"
@@ -122,7 +124,6 @@ read_packet(int sock,RecoveryPacket * packet)
 	char * func = "read_packet()";
 #endif			
 	int r = 0;
-	int cnt = 0;
 	char * read_ptr = NULL;
 	int read_size = 0;
 	int packet_size = 0;
@@ -138,33 +139,18 @@ read_packet(int sock,RecoveryPacket * packet)
 		r = recv(sock,read_ptr + read_size ,packet_size - read_size, MSG_WAITALL);
 		if (r < 0)
 		{
-			if (errno == EINTR)
-			{
-				usleep(PGR_RECV_WAIT_MSEC);
+			if (errno == EINTR || errno == EAGAIN)
 				continue;
-			}
-#ifdef EAGAIN
-			if (errno == EAGAIN)
-			{
-				usleep(PGR_RECV_WAIT_MSEC);
-				continue;
-			}
-#endif /* EAGAIN */
-			if (cnt < PGR_RECV_RETRY_CNT )
-			{
-				cnt ++;
-				usleep(PGR_RECV_WAIT_MSEC);
-				continue;
-			}
 			else
 			{
+				show_error("%s:recv failed: %d(%s)",func, errno, strerror(errno));
 				return -1;
 			}
 		}
-		if (r > 0)
+		else if (r > 0)
 		{
 			read_size += r;
-			if (read_size == PGR_MESSAGE_BUFSIZE)
+			if (read_size == packet_size)
 			{
 #ifdef PRINT_DEBUG
 				show_debug("%s:receive packet",func);
@@ -173,20 +159,13 @@ read_packet(int sock,RecoveryPacket * packet)
 				return read_size;
 			}
 		}
-		return -1;
-		/*
-		if (cnt < PGR_RECV_RETRY_CNT )
-		{
-			cnt ++;
-			usleep(PGR_RECV_WAIT_MSEC);
-			continue;
-		}
-		else
+		else /* r == 0 */
 		{
+			show_error("%s:unexpected EOF", func);
 			return -1;
 		}
-		*/
 	}
+	return -1;
 }
 
 static int
@@ -197,17 +176,23 @@ read_packet_from_master( RecoveryTbl * host, RecoveryPacket * packet )
 	fd_set	  rmask;
 	struct timeval timeout;
 
-	timeout.tv_sec = PGR_RECV_TIMEOUT;
-	timeout.tv_usec = 0;
-
-	/*
-	 * Wait for something to happen.
-	 */
-	FD_ZERO(&rmask);
-	FD_SET(host->recovery_sock,&rmask);
 	for(;;)
 	{
+		timeout.tv_sec = PGR_RECV_TIMEOUT;
+		timeout.tv_usec = 0;
+
+		/*
+		 * Wait for something to happen.
+		 */
+		FD_ZERO(&rmask);
+		FD_SET(host->recovery_sock,&rmask);
 		rtn = select(host->recovery_sock+1, &rmask, (fd_set *)NULL, (fd_set *)NULL, &timeout);
+		
+		if (rtn == 0) /* timeout */
+		{
+			return -1;
+		}
+
 		if (rtn && FD_ISSET(host->recovery_sock, &rmask))
 		{
 			read_size = read_packet(host->recovery_sock, packet);
@@ -219,6 +204,7 @@ read_packet_from_master( RecoveryTbl * host, RecoveryPacket * packet )
 static int
 send_recovery_packet(int  sock, RecoveryPacket * packet)
 {
+	char *func = "send_recovery_packet";
 	char * send_ptr;
 	int send_size= 0;
 	int buf_size = 0;
@@ -230,18 +216,23 @@ send_recovery_packet(int  sock, RecoveryPacket * packet)
 	for (;;)
 	{
 		s = send(sock, send_ptr + send_size,buf_size - send_size ,0);
-		if (s < 0){
-			if (errno == EINTR)
-			{
+		if (s < 0)
+		{
+			if (errno == EINTR || errno == EAGAIN)
 				continue;
-			}
+
+			show_error("%s:send error: %d(%s)", func, errno, strerror(errno));
 			return STATUS_ERROR;
 		}
-		if (s == 0)
+		else if (s == 0)
 		{
-			return STATUS_OK;
+			show_error("%s:unexpected EOF", func);
+			return STATUS_ERROR;
 		}
+
 		send_size += s;
+		if (send_size == buf_size)
+			return STATUS_OK;
 	}
 }
 
@@ -257,7 +248,7 @@ send_packet(RecoveryTbl * host, RecoveryPacket * packet )
 		{
 			if (count > MAX_RETRY_TIMES )
 			{
-				show_error("%s:PGR_Create_Socket_Connect failed",func);
+				show_error("%s:host[%s] port[%d]PGR_Create_Socket_Connect failed",func,host->hostName, host->recoveryPort);
 				return STATUS_ERROR;
 			}
 			count ++;
@@ -328,22 +319,15 @@ PGRsend_load_balance_packet(RecoveryPacket * packet)
 			close(lbp->recovery_sock);
 			lbp->recovery_sock = -1;
 		}
+#ifdef PRINT_DEBUG
+	show_debug("%s:host[%s] port[%d]",func,lbp->hostName,lbp->recoveryPort);
+#endif
 		status = send_packet(lbp,packet);
 		if (lbp->recovery_sock != -1)
 		{
 			close(lbp->recovery_sock);
 			lbp->recovery_sock = -1;
 		}
-		/*
-		if (status == STATUS_OK)
-		{
-			if (lbp->sock != -1)
-			{
-				close(lbp->sock);
-				lbp->sock = -1;
-			}
-		}
-		*/
 		lbp ++;
 	}
 	return STATUS_OK;
@@ -373,7 +357,8 @@ send_vacuum(HostTbl *host, char * userName, int stage)
 	header.cmdType = CMD_TYPE_VACUUM;
 	header.pid = getpid();
 	header.query_id = getpid();
-	rtn = PGRsend_replicate_packet_to_server(host,&header,query,PGR_Result);
+	header.isAutoCommit=1;
+	rtn = PGRsend_replicate_packet_to_server(host,&header,query,PGR_Result,0, true);
 	if (query !=NULL)
 		free(query);
 	return rtn;	
@@ -448,7 +433,7 @@ retry_connect_master:
 		 * connection error , master may be down
 		 */
 		show_error("%s:get master info error , master may be down",func);
-		PGRset_recovery_packet_no(packet, RECOVERY_ERROR_CONNECTION);
+		PGRset_recovery_packet_no(packet, RECOVERY_ERROR_TARGET_ONLY);
 		status = send_packet(&Target, packet);
 		reset_recovery_prepare();
 		loop_end = true;
@@ -460,14 +445,8 @@ retry_connect_master:
 	status = send_vacuum(master, userName, PGR_RECOVERY_1ST_STAGE );
 	if (status != STATUS_OK)
 	{
-		show_error("%s:vacuum error , master may be down",func);
-		PGRset_recovery_packet_no(packet, RECOVERY_ERROR_CONNECTION);
-		status = send_packet(&Target, packet);
-		reset_recovery_prepare();
-		loop_end = true;
-		if (userName != NULL)
-			free(userName);
-		return loop_end;
+		PGRset_host_status(master, DB_TBL_ERROR);
+		goto retry_connect_master;
 	}
 
 	memcpy(Master.hostName,master->hostName,sizeof(Master.hostName));
@@ -525,7 +504,7 @@ retry_connect_master:
 		if (status != STATUS_OK)
 		{
 			show_error("%s:no[%d] send_packet to target error",func,ntohs(packet->packet_no));
-			PGRset_recovery_packet_no(packet, RECOVERY_ERROR_CONNECTION);
+			PGRset_recovery_packet_no(packet, RECOVERY_ERROR_TARGET_ONLY);
 			status = send_packet(&Master,packet);
 			reset_recovery_prepare();
 			loop_end = true;
@@ -533,6 +512,8 @@ retry_connect_master:
 	}
 	if (userName != NULL)
 		free(userName);
+
+
 	return loop_end;
 }
 
@@ -546,36 +527,59 @@ second_setup_recovery (RecoveryPacket * packet)
 	char * userName = NULL;
 
 	start_recovery();
+
+	/* send vauum command to master server */
+	while ((master = PGRget_master()) != NULL)
+	{
+		userName = strdup(packet->userName);
+		status = send_vacuum(master, userName, PGR_RECOVERY_2ND_STAGE );
+		if (status != STATUS_OK)
+		{
+			PGRset_host_status(master, DB_TBL_ERROR);
+			if (userName != NULL)
+			{
+				free(userName);
+				userName = NULL;
+			}
+			continue;
+		}
+		break;
+	}
+
+	if (master == NULL)
+	{
+		show_error("%s:vacuum error , master may be down",func);
+		PGRset_recovery_packet_no(packet, RECOVERY_ERROR_TARGET_ONLY);
+		status = send_packet(&Target,packet);
+		status = send_packet(&Master,packet);
+		reset_recovery_prepare();
+
+		return true;		
+	}
+
 	/*
 	 * wait until all started transactions are going to finish
 	 */
-	status = PGRwait_transaction_count_clear();
-	if (status != STATUS_OK)
+	status = PGRwait_transaction_count_clear(master);
+	if (status != STATUS_OK && PGRget_recovery_status() == RECOVERY_START)
 	{
 		show_error("%s:transaction is too busy, please try again after",func);
-		PGRset_recovery_packet_no(packet, RECOVERY_ERROR_CONNECTION);
+		PGRset_recovery_packet_no(packet, RECOVERY_ERROR_TARGET_ONLY);
 		status = send_packet(&Target,packet);
 		status = send_packet(&Master,packet);
 		reset_recovery_prepare();
-		loop_end = true;
-		return loop_end;
+		return true;
 	}
 
-	/* send vauum command to master server */
-	master = PGRget_master();
-	userName = strdup(packet->userName);
-	status = send_vacuum(master, userName, PGR_RECOVERY_2ND_STAGE );
-	if (status != STATUS_OK)
-	{
-		show_error("%s:vacuum error , master may be down",func);
+	if (PGRget_recovery_status() != RECOVERY_START) {
+		show_error("%s:queue set failed. stop to recovery",func);
 		PGRset_recovery_packet_no(packet, RECOVERY_ERROR_CONNECTION);
 		status = send_packet(&Target,packet);
 		status = send_packet(&Master,packet);
 		reset_recovery_prepare();
-		loop_end = true;
 		if (userName != NULL)
 			free(userName);
-		return loop_end;
+		return true;
 	}
 
 	/*
@@ -593,11 +597,22 @@ second_setup_recovery (RecoveryPacket * packet)
 		status = send_packet(&Target,packet);
 		status = send_packet(&Master,packet);
 		reset_recovery_prepare();
-		loop_end = true;
 		if (userName != NULL)
 			free(userName);
-		return loop_end;
+		return true;
 	}
+
+	if (PGRget_recovery_status() != RECOVERY_START) {
+         	show_error("%s:queue set failed. stop to recovery",func);
+		PGRset_recovery_packet_no(packet, RECOVERY_ERROR_CONNECTION);
+		status = send_packet(&Target,packet);
+		status = send_packet(&Master,packet);
+		reset_recovery_prepare();
+		if (userName != NULL)
+			free(userName);
+		return true;
+	}
+
 	/*
 	 * wait answer from master server 
 	 */
@@ -619,7 +634,7 @@ second_setup_recovery (RecoveryPacket * packet)
 	}
 	else
 	{
-		show_error("%s:failer answer returned",func);
+		show_error("%s:failure answer returned",func);
 		PGRset_recovery_packet_no(packet, RECOVERY_ERROR_CONNECTION);
 		status = send_packet(&Target,packet);
 		status = send_packet(&Master,packet);
@@ -630,7 +645,13 @@ second_setup_recovery (RecoveryPacket * packet)
 		free(userName);
 	return loop_end;
 }
-
+/**
+ * send queries from queue.
+ *
+ * return
+ *   STATUS_OK - success both
+ *   STATUS_ERROR - fail both
+ */
 static int
 PGRsend_queue(RecoveryTbl * master, RecoveryTbl * target)
 {
@@ -646,7 +667,7 @@ PGRsend_queue(RecoveryTbl * master, RecoveryTbl * target)
 	long mtype = 0;
 	int status = 0;
 	int query_size = 0;
-
+	int rtn=0;
 	if (master == (RecoveryTbl *)NULL)
 	{
 		show_error("%s:there is no master ",func);
@@ -684,7 +705,7 @@ PGRsend_queue(RecoveryTbl * master, RecoveryTbl * target)
 	memset(msg_header,0,size+4);
 	rcv_header_size = sizeof(long)+sizeof(ReplicateHeader);
 	status = STATUS_OK;
-	while (msgrcv(RecoveryMsgid , msg_header, rcv_header_size, RECOVERY_HEADER_MTYPE, IPC_NOWAIT) > 0 )
+	while (msgrcv(*RecoveryMsgid , msg_header, rcv_header_size, RECOVERY_HEADER_MTYPE, IPC_NOWAIT) > 0 )
 	{
 		header = (ReplicateHeader *)(msg_header->mdata);
 		query_size = ntohl(header->query_size);
@@ -708,7 +729,7 @@ PGRsend_queue(RecoveryTbl * master, RecoveryTbl * target)
 			return STATUS_ERROR;
 		}
 		memset(msg_query,0,size+4);
-		if (msgrcv(RecoveryMsgid , msg_query, sizeof(long)+query_size, mtype, IPC_NOWAIT) < 0)
+		if (msgrcv(*RecoveryMsgid , msg_query, sizeof(unsigned int)+query_size, mtype, IPC_NOWAIT) < 0)
 		{
 			show_error("%s:msgrcv() failed. (%s)",func,strerror(errno) );
 			status = STATUS_ERROR;
@@ -720,10 +741,12 @@ PGRsend_queue(RecoveryTbl * master, RecoveryTbl * target)
 			break;
 		}
 		query = (char *)(msg_query->mdata);
-		if (PGRsend_replicate_packet_to_server(master_ptr,header,query,PGR_Result) != STATUS_OK)
+		show_debug("%s:put query from queue :%s",func,query);
+		rtn=PGRsend_replicate_packet_to_server(master_ptr,header,query,PGR_Result,msg_query->replicationId, true);
+		if (rtn == STATUS_ERROR || rtn == STATUS_DEADLOCK_DETECT )
 		{
 			show_error("%s:PGRsend_replicate_packet_to_server to master error",func);
-			status = STATUS_ERROR;
+			status = STATUS_ERROR ;
 			if (msg_query != NULL)
 			{
 				free(msg_query);
@@ -733,7 +756,8 @@ PGRsend_queue(RecoveryTbl * master, RecoveryTbl * target)
 		}
 		if (target_ptr != NULL)
 		{
-			if (PGRsend_replicate_packet_to_server(target_ptr,header,query,PGR_Result) != STATUS_OK)
+			rtn=PGRsend_replicate_packet_to_server(target_ptr,header,query,PGR_Result,msg_query->replicationId, true);
+			if (rtn == STATUS_ERROR || rtn == STATUS_DEADLOCK_DETECT )
 			{
 				show_error("%s:PGRsend_replicate_packet_to_server to target error",func);
 				status = STATUS_ERROR;
@@ -772,7 +796,14 @@ pgrecovery_loop(int fd)
 	bool loop_end = false;
 	RecoveryPacket packet;
 	HostTbl new_host;
+	RecoveryTbl * lbp;
 
+	lbp = LoadBalanceTbl;
+	if (lbp == (RecoveryTbl *)NULL)
+	{
+		show_error("%s:recovery table is NULL",func);
+		return ;
+	}
 #ifdef PRINT_DEBUG
 	show_debug("%s:recovery accept port %d",func, Recovery_Port_Number);
 #endif			
@@ -788,6 +819,10 @@ pgrecovery_loop(int fd)
 		}
 		count ++;
 	}
+	if(sock==-1) {
+			show_error("can't create recovery socket.exit.");
+			PGRreplicate_exit(1);
+	}
 	for(;;)
 	{
 		int read_size = 0;
@@ -807,6 +842,12 @@ pgrecovery_loop(int fd)
 		 * read packet from target cluster server
 		 */
 		rtn = select(sock+1, &rmask, (fd_set *)NULL, (fd_set *)NULL, &timeout);
+
+		if (rtn == 0) /* timeout */
+		{
+			return;
+		}
+
 		if (rtn && FD_ISSET(sock, &rmask))
 		{
 			read_size = read_packet(sock, &packet);
@@ -826,13 +867,16 @@ pgrecovery_loop(int fd)
 				/*
 				 * start prepare of recovery
 				 */
-				loop_end = first_setup_recovery(&sock, &packet);
+
 #ifdef PRINT_DEBUG
 				show_debug("%s:1st master %s - %d",
 					func,Master.hostName,Master.port);
 				show_debug("%s:1st target %s - %d",
 					func,Target.hostName,Target.port);
 #endif			
+
+				loop_end = first_setup_recovery(&sock, &packet);
+				show_debug("%s:first_setup_recovery end:%d ",func,loop_end);
 				break;
 			case RECOVERY_START_REQ : 
 				/*
@@ -866,7 +910,6 @@ pgrecovery_loop(int fd)
 				show_debug("%s:last target %s - %d",
 					func, Target.hostName,Target.port);
 #endif			
-				sleep(3);
 				status = PGRsend_queue(&Master,&Target);
 				if (status == STATUS_OK)
 				{
@@ -884,7 +927,7 @@ pgrecovery_loop(int fd)
 					status = send_packet(&Target,&packet);
 				}
 				/*
-				 * stop queueing, and initiarise recovery status
+				 * stop queueing, and re-initialize recovery status
 				 */
 				finish_recovery();
 				loop_end = true;
@@ -896,6 +939,11 @@ pgrecovery_loop(int fd)
 					close(Master.recovery_sock);
 					Master.recovery_sock = -1;
 				}
+				if (Target.recovery_sock != -1)
+				{
+					close(Target.recovery_sock);
+					Target.recovery_sock = -1;
+				}
 				send_packet(&Master, &packet);
 				MasterPacketData.packet_no = packet.packet_no;
 				PGRsend_load_balance_packet(&MasterPacketData);
@@ -916,6 +964,10 @@ pgrecovery_loop(int fd)
 				PGRsend_load_balance_packet(&MasterPacketData);
 				memset((char *)&MasterPacketData,0,sizeof(RecoveryPacket));
 				break;
+		default:
+		  show_error("%s:unknown packet. abort to parse");
+			     loop_end=true;
+			     break;
 		}
 		if (loop_end)
 		{
